# By Remi Vuillemot

import numpy as np
import copy
from Bio.SVDSuperimposer import SVDSuperimposer

class ContinuousFlexPDBHandler:
    
    @classmethod
    def read_coords(cls, pdb_file):
        print("> Reading pdb file %s ..." % pdb_file)
        coords = []
        with open(pdb_file, "r") as f:
            for line in f:
                if 'ATOM' in line:
                    coords.append([
                            line[30:38],line[38:46],line[46:54]
                        ])
        return np.array(coords).astype(float)

    def __init__(self, pdb_file):
        """
        Contructor
        :param pdb_file: PDB file
        """
        atom = []
        atomNum = []
        atomName = []
        resName = []
        resAlter = []
        chainName = []
        resNum = []
        coords = []
        occ = []
        temp = []
        chainID = []
        elemName = []
        print("> Reading pdb file %s ..." % pdb_file)
        with open(pdb_file, "r") as f:
            for line in f:
                spl = line.split()
                if len(spl) > 0:
                    if (spl[0] == 'ATOM'):  # or (hetatm and spl[0] == 'HETATM'):
                        l = [line[:6], line[6:11], line[12:16], line[16], line[17:21], line[21], line[22:26],
                             line[30:38],
                             line[38:46], line[46:54], line[54:60], line[60:66], line[72:76], line[76:78]]
                        l = [i.strip() for i in l]
                        atom.append(l[0])
                        atomNum.append(l[1])
                        atomName.append(l[2])
                        resAlter.append(l[3])
                        resName.append(l[4])
                        chainName.append(l[5])
                        resNum.append(l[6])
                        coords.append([float(l[7]), float(l[8]), float(l[9])])
                        occ.append(l[10])
                        temp.append(l[11])
                        chainID.append(l[12])
                        elemName.append(l[13])

        atomNum = np.array(atomNum)
        atomNum[np.where(atomNum == "*****")[0]] = "-1"

        self.atom = np.array(atom, dtype='<U6')
        self.n_atoms = len(self.atom)
        self.atomNum = np.array(atomNum).astype(int)
        self.atomName = np.array(atomName, dtype='<U4')
        self.resName = np.array(resName, dtype='<U4')
        self.resAlter = np.array(resAlter, dtype='<U1')
        self.chainName = np.array(chainName, dtype='<U1')
        self.resNum = np.array(resNum).astype(int)
        self.coords = np.array(coords).astype(float)
        self.occ = np.array(occ).astype(float)
        self.temp = np.array(temp).astype(float)
        self.chainID = np.array(chainID, dtype='<U4')
        self.elemName = np.array(elemName, dtype='<U2')

        if self.n_atoms == 0 :
            raise RuntimeError("Could not read PDB file : PDB file is empty")

        print("\t Done \n")

    def write_pdb(self, file):
        """
        Write to PDB Format
        :param file: pdb file path
        """
        print("> Writing pdb file %s ..." % file)
        with open(file, "w") as file:
            past_chainName = self.chainName[0]
            past_chainID = self.chainID[0]
            for i in range(len(self.atom)):
                if past_chainName != self.chainName[i] or past_chainID != self.chainID[i]:
                    past_chainName = self.chainName[i]
                    past_chainID = self.chainID[i]
                    file.write("TER\n")

                atom = self.atom[i].ljust(6)  # atom#6s
                if self.atomNum[i] == -1 or self.atomNum[i] >= 100000:
                    atomNum = "99999"  # aomnum#5d
                else:
                    atomNum = str(self.atomNum[i]).rjust(5)  # aomnum#5d
                atomName = self.atomName[i].ljust(4)  # atomname$#4s
                resAlter = self.resAlter[i].ljust(1)  # resAlter#1
                resName = self.resName[i].ljust(4)  # resname#1s
                chainName = self.chainName[i].rjust(1)  # Astring
                resNum = str(self.resNum[i]).rjust(4)  # resnum
                coordx = str('%8.3f' % (float(self.coords[i][0]))).rjust(8)  # x
                coordy = str('%8.3f' % (float(self.coords[i][1]))).rjust(8)  # y
                coordz = str('%8.3f' % (float(self.coords[i][2]))).rjust(8)  # z\
                occ = str('%6.2f' % self.occ[i]).rjust(6)  # occ
                temp = str('%6.2f' % self.temp[i]).rjust(6)  # temp
                chainID = str(self.chainID[i]).ljust(4)  # elname
                elemName = str(self.elemName[i]).rjust(2)  # elname
                file.write("%s%s %s%s%s%s%s    %s%s%s%s%s      %s%s\n" % (
                atom, atomNum, atomName, resAlter, resName, chainName, resNum,
                coordx, coordy, coordz, occ, temp, chainID, elemName))
            file.write("END\n")
        print("\t Done \n")

    def matchPDBatoms(self, reference_pdb, ca_only=False, matchingType=None):
        """
        match atoms between the pdb and a reference pdb
        :param reference_pdb: ContinuousflexPDBHandler
        :param ca_only: True if carbon alph only
        :param matchingType: 0= chain first, 1= segment ID first
        :return: index of matching atoms
        """
        print("> Matching PDBs atoms ...")
        n_mols = 2

        if matchingType == None:
            chain_name_list1 = self.get_chain_list(chainType=0)
            chain_name_list2 = reference_pdb.get_chain_list(chainType=0)
            n_matching_chain_names = sum([i in chain_name_list2 for i in chain_name_list1])

            chain_id_list1 = self.get_chain_list(chainType=1)
            chain_id_list2 = reference_pdb.get_chain_list(chainType=1)
            n_matching_chain_ids = sum([i in chain_id_list2 for i in chain_id_list1])

            if n_matching_chain_ids >n_matching_chain_names:
                matchingType = 1
                print("\t Matching segments %s ... "%n_matching_chain_ids)
            elif  n_matching_chain_ids < n_matching_chain_names:
                matchingType = 0
                print("\t Matching chains %s ... "%n_matching_chain_names)
            else:
                raise RuntimeError("No matching chains")


        ids = []
        ids_idx = []
        for m in [self, reference_pdb]:
            id_tmp = []
            id_idx_tmp = []
            for i in range(m.n_atoms):
                if (not ca_only) or m.atomName[i] == "CA" or m.atomName[i] == "P":
                    id_tmp.append("%s_%i_%s_%s" % (m.chainName[i] if matchingType == 0 else m.chainID[i],
                                                   m.resNum[i], m.resName[i], m.atomName[i]))
                    id_idx_tmp.append(i)
            ids.append(np.array(id_tmp))
            ids_idx.append(np.array(id_idx_tmp))

        idx = []
        for i in range(len(ids[0])):
            idx_line = [ids_idx[0][i]]
            for m in range(1, n_mols):
                idx_tmp = np.where(ids[0][i] == ids[m])[0]
                if len(idx_tmp) == 1:
                    idx_line.append(ids_idx[m][idx_tmp[0]])
                elif len(idx_tmp) > 1:
                    print("\t Warning : One atom in mol#0 is matching several atoms in mol#%i : " % m)

            if len(idx_line) == n_mols:
                idx.append(idx_line)

        if len(idx) == 0:
            print("\t Warning : No matching coordinates")

        print("\t %i matching atoms " % len(np.array(idx)))
        print("\t Done")

        return np.array(idx)

    def alignMol(self, reference_pdb, idx_matching_atoms=None):
        print("> Aligning PDB ...")

        if idx_matching_atoms is not None:
            c1 = reference_pdb.coords[idx_matching_atoms[:, 1]]
            c2 = self.coords[idx_matching_atoms[:, 0]]
        else:
            c1 = reference_pdb.coords
            c2 = self.coords

        rot, tran = self.alignCoords(c1,c2)
        self_copy = self.copy()
        self_copy.coords = np.dot(self_copy.coords, rot) + tran
        print("\t Done \n")

        return self_copy

    @classmethod
    def alignCoords(cls, coord_ref, coord):
        sup = SVDSuperimposer()
        sup.set(coord_ref, coord)
        try:
            sup.run()
            rot, tran = sup.get_rotran()
        except np.linalg.LinAlgError:
            print("Error while aligning : SVD did not converge")
            rot = np.eye(3)
            tran = np.zeros(3)
        return rot, tran

    def getRMSD(self, reference_pdb, align=False, idx_matching_atoms=None):
        if align:
            aligned = self.alignMol(reference_pdb=reference_pdb, idx_matching_atoms=idx_matching_atoms)
        else:
            aligned=self
        if idx_matching_atoms is not None:
            coord1 = reference_pdb.coords[idx_matching_atoms[:, 1]]
            coord2 = aligned.coords[idx_matching_atoms[:, 0]]
        else:
            coord1 = reference_pdb.coords
            coord2 = aligned.coords
        return np.sqrt(np.mean(np.square(np.linalg.norm(coord1 - coord2, axis=1))))

    def select_atoms(self, idx):
        self.coords = self.coords[idx]
        self.n_atoms = self.coords.shape[0]
        self.atom = self.atom[idx]
        self.atomNum = self.atomNum[idx]
        self.atomName = self.atomName[idx]
        self.resName = self.resName[idx]
        self.resAlter = self.resAlter[idx]
        self.chainName = self.chainName[idx]
        self.resNum = self.resNum[idx]
        self.elemName = self.elemName[idx]
        self.occ = self.occ[idx]
        self.temp = self.temp[idx]
        self.chainID = self.chainID[idx]

    def get_chain_list(self, chainType=0):
        if chainType == 0:
            lst = list(set(self.chainName))
        else:
            lst = list(set(self.chainID))
        lst.sort()
        return lst

    def get_chain(self, chainName):
        if not isinstance(chainName, list):
            chainName=[chainName]
        chainidx =[]
        for i in chainName:
            idx = np.where(self.chainName == i)[0]
            if len(idx) == 0:
                idx= np.where(self.chainID == i)[0]
            chainidx = chainidx + list(idx)
        return np.array(chainidx)

    def select_chain(self, chainName):
        self.select_atoms(self.get_chain(chainName))

    def copy(self):
        return copy.deepcopy(self)

    def remove_alter_atom(self):
        idx = []
        for i in range(self.n_atoms):
            if self.resAlter[i] != "":
                print("!!! Alter residue %s for atom %i"%(self.resName[i], self.atomNum[i]))
                if self.resAlter[i] == "A":
                    idx.append(i)
                    self.resAlter[i]=""
            else:
                idx.append(i)
        self.select_atoms(idx)

    def remove_hydrogens(self):
        idx=[]
        for i in range(self.n_atoms):
            if not self.atomName[i].startswith("H"):
                idx.append(i)
        self.select_atoms(idx)

    def alias_atom(self, atomName, atomNew, resName=None):
        n_alias = 0
        for i in range(self.n_atoms):
            if self.atomName[i] == atomName:
                if resName is not None :
                    if self.resName[i] == resName :
                        self.atomName[i] = atomNew
                        n_alias+=1
                else:
                    self.atomName[i] = atomNew
                    n_alias+=1
        print("%s -> %s : %i lines changed"%(atomName, atomNew, n_alias))

    def alias_res(self, resName, resNew):
        n_alias=0
        for i in range(self.n_atoms):
            if self.resName[i] == resName :
                self.resName[i] = resNew
                n_alias+=1
        print("%s -> %s : %i lines changed"%(resName ,resNew,  n_alias))


    def add_terminal_res(self):
        aa = ["ALA", "CYS", "ASP", "GLU", "PHE", "GLY", "HIS", "ILE", "LYS", "LEU", "MET", "ASN", "PRO",
              "GLN", "ARG", "SER", "THR", "VAL", "TRP", "TYR"]
        past_chainName = self.chainName[0]
        past_chainID = self.chainID[0]
        for i in range(self.n_atoms-1):
            if past_chainName != self.chainName[i+1] or past_chainID != self.chainID[i+1]:
                if self.resName[i] in aa :
                    print("End of chain %s ; adding terminal residue to %s %i %s"%
                          (past_chainID,self.resName[i],self.resNum[i],self.atomName[i]))
                    resNum = self.resNum[i]
                    j=0
                    while self.resNum[i-j] ==resNum :
                        self.resName[i - j] += "T"
                        j+=1
                else:
                    print("End of chain %s %s %i"% (past_chainID,self.resName[i],self.resNum[i]))
                past_chainName = self.chainName[i+1]
                past_chainID = self.chainID[i+1]


        i = self.n_atoms-1
        if self.resName[i] in aa:
            print("End of chain %s ; adding terminal residue to %s %i %s" % (
            past_chainID, self.resName[i], self.resNum[i], self.atomName[i]))
            resNum = self.resNum[i]
            j = 0
            while self.resNum[i - j] == resNum:
                self.resName[i - j] += "T"
                j += 1
        else:
            print("End of chain %s %s %i" % (past_chainID, self.resName[i], self.resNum[i]))


    def check_res_order(self):
        chains = list(set(self.chainID))
        chains.sort()
        new_idx = []
        for c in chains:
            chain_idx = self.get_chain(c)
            resNumlist = list(set(self.resNum[chain_idx]))
            resNumlist.sort()
            for i in range(len(resNumlist)):
                idx = np.where(self.resNum[chain_idx] == resNumlist[i])[0]
                new_idx += list(chain_idx[idx])
        self.select_atoms(np.array(new_idx))

    def atom_res_reorder(self):
        chains = list(set(self.chainID))
        chains.sort()

        # reorder atoms and res
        for c in chains:
            chain_idx = self.get_chain(c)
            past_resNum = self.resNum[chain_idx[0]]
            resNum = 1
            for i in range(len(chain_idx)):
                if self.resNum[chain_idx[i]] != past_resNum:
                    if self.resNum[chain_idx[i]] != past_resNum+1:
                        print("ERROR : non sequential residue number in one segment")
                    past_resNum = self.resNum[chain_idx[i]]
                    resNum += 1
                self.resNum[chain_idx[i]] = resNum
                self.atomNum[chain_idx[i]] = i + 1

    def allatoms2ca(self):
        new_idx = []
        for i in range(self.n_atoms):
            if self.atomName[i] == "CA" or self.atomName[i] == "P":
                new_idx.append(i)
        return np.array(new_idx)

    def center(self):
        self.coords -= np.mean(self.coords, axis=0)




from .test_workflow_HEMNMA import *
from .test_workflow_HEMNMA3D import *
from .test_workflow_StA import *
from .test_structure_mapping import *
from .test_workflow_subtomogram_synthesize import *
from .test_workflow_TomoFlow import *
from .test_workflow_GENESIS import *
from pyworkflow.tests import DataSet

files_dictionary = {'pdb': 'pdb/AK.pdb', 'particles': 'particles/img.stk', 'vol': 'volumes/AK_LP10.vol',
                    'precomputed_atomic': 'gold/images_WS_atoms.xmd',
                    'precomputed_pseudoatomic': 'gold/images_WS_pseudoatoms.xmd',
                    'small_stk': 'test_alignment_10images/particles/smallstack_img.stk',
                    'subtomograms':'HEMNMA_3D/subtomograms/*.vol',
                    'precomputed_HEMNMA3D_atoms':'HEMNMA_3D/gold/precomputed_atomic.xmd',
                    'precomputed_HEMNMA3D_pseudo':'HEMNMA_3D/gold/precomputed_pseudo.xmd',

                    'charmm_prm':'genesis/par_all36_prot.prm',
                    'charmm_top':'genesis/top_all36_prot.rtf',
                    '1ake_pdb':'genesis/1ake.pdb',
                    '1ake_vol':'genesis/1ake.mrc',
                    '4ake_pdb':'genesis/4ake.pdb',
                    '4ake_aa_pdb':'genesis/4ake_aa.pdb',
                    '4ake_aa_psf':'genesis/4ake_aa.psf',
                    '4ake_ca_pdb':'genesis/4ake_ca.pdb',
                    '4ake_ca_top':'genesis/4ake_ca.top',
                    }
DataSet(name='nma_V2.0', folder='nma_V2.0', files=files_dictionary,
        url='https://raw.githubusercontent.com/continuousflex-org/testdata-continuousflex/main')



"""
This file reads, cleans an 'xmd' file and 
create a dataframe and a numpy array that
contains only the normal modes amplitudes
"""

import re
import numpy as np
from math import cos, sin, radians
from .euler2quaternion import eul2quat
import torch

def header(path):
    num_chars = 20
    head = 0
    with open(path, 'r') as f:
        lines = f.readlines()
        for line in lines:
            if len(line) < num_chars:
                head+=1
            else:
                pass
    return head

def read_file(path):
    head = header(path)
    f = open(path, 'r')
    file_list = f.readlines()
    column_names = [col.replace('\n', '') for col in file_list[4:head]]
    for i in range(head):
        file_list.pop(0)
    return file_list, column_names


def create_array(path, flag='nma'):
    file_list, column_names = read_file(path)
    columns = len(list(filter(None,re.split("\s|'", file_list[0]))))
    num_modes = columns-(len(column_names)-1)
    nma_index = column_names.index(' _nmaDisplacements')
    for i in range(num_modes):
        column_names.insert(nma_index, 'mode '+str(num_modes - i))
    column_names.remove(' _nmaDisplacements')
    img_index = column_names.index(' _image')
    rot_index = column_names.index(' _angleRot')
    tilt_index = column_names.index(' _angleTilt')
    psi_index = column_names.index(' _anglePsi')
    shiftx_index = column_names.index(' _shiftX')
    shifty_index = column_names.index(' _shiftY')
    
    for i in range(len(file_list)):
        file_list[i] = file_list[i].replace('\n', ' ')
        file_list[i] = list(filter(None, re.split("\s|'",file_list[i])))
    
    data_array=np.reshape(file_list,(len(file_list),columns))
    img_names=data_array[:,img_index]
    nm_amplitudes = data_array[:, nma_index: nma_index+num_modes].astype('float32') 
    nm_amplitudes, nma_min, nma_max = min_max(nm_amplitudes)
    angles = data_array[:, [rot_index, tilt_index, psi_index]].astype('float32')
    shifts = data_array[:, [shiftx_index, shifty_index]].astype('float32')
    shifts, shf_min, shf_max = min_max(shifts)
    quaternions = np.zeros((angles.shape[0], 4), dtype='float32')
    for i in range(len(angles)):
        quaternions[i,:] = eul2quat(angles, i)
    if flag=='nma':
        print("Number of Normal Modes detected is: ",num_modes)
        return nm_amplitudes, nma_min, nma_max, img_names
    elif flag=='ang':
        return quaternions, img_names
    elif flag=='shf':
        return shifts, shf_min, shf_max, img_names
    else:
        raise ValueError('Unknown flag, you must select nma for Normal mode amplitudes, ang for euler angles, shf for shifts (X and Y)')


def min_max(arr):
    _min = torch.min(arr, dim=0)
    _max = torch.max(arr, dim=0)
    arr = (arr - _min[0]) / (_max[0] - _min[0])
    return arr, _min[0], _max[0]


def standardization(arr, params: bool = False, num_modes: int = 3):
    _mean = []
    _mu = []
    num_params = 0
    if params:
        num_params = num_modes + 5
    else:
        num_params = num_modes
    for i in range(num_params):
        _mean.append(np.mean(arr[:, i]))
        _mu.append(np.std(arr[:, i]))
    for i in range(num_params):
        for j in range(len(arr)):
            tmp = arr[j, i]
            arr[j, i] = (tmp-_mean[i])/_mu[i]
    return arr, _mean, _mu

def reverse_min_max(arr, _min, _max):
    """
    This function rescale back the target values to its original range
    it rescaled it back and put it in a list then reshape it to an array
    of the same shape as the input
    Parameters
    ----------
    arr : numpy array float32
        a numpy array for example (100,3).

    Returns
    -------
    rescaled_output : numpy array float32
        rescaled_output: a numpy array of the same shape as input for
        example (100, 3).

    """
    return (arr * (_max - _min)) + _min

def reverse_standardization(arr, _mean, _mu, params: bool = False, num_modes:int =3):
    num_params = 0
    if params:
        num_params = num_modes + 5
    else:
        num_params = num_modes
    rescaled_list = []
    for i in range(len(arr)):
        for j in range(params):
            rescaled_list.append((arr[i,j]*_mu[j])+_mean[j])
    rescaled_output = np.array(rescaled_list).reshape((len(arr), num_params))
    return rescaled_output



def rotation_matrix(euler_angles):
    rot = radians(euler_angles[0])
    tilt = radians(euler_angles[1])
    psi = radians(euler_angles[2])
    rot_mat = np.array([[(cos(rot)*cos(tilt)*cos(psi))-(sin(rot)*sin(psi)),
                         (-cos(psi)*sin(rot))-(cos(rot)*cos(tilt)*sin(psi)), 
                         (cos(rot)*sin(tilt))],
                        [(cos(rot)*sin(psi))+(cos(tilt)*cos(psi)*sin(rot)),
                         (cos(rot)*cos(psi))-(cos(tilt)*sin(rot)*sin(psi)),
                          (sin(rot)*sin(tilt))],
                        [-cos(psi)*sin(tilt), sin(psi)*sin(tilt),cos(tilt)]])

    return rot_mat



# Author: Ilyes Hamitouche

import torch.nn as nn
from torchvision import transforms
import torch.optim as optim
from torch.utils.data import DataLoader
from processing_dh.data import cryodata
from processing_dh.models import deephemnma
import numpy as np
import torch
from torch.utils.data.sampler import SubsetRandomSampler
from torch.utils.tensorboard import SummaryWriter
import sys
def norm(imgs_path, output_path, FLAG, mode, batch_size):
    dataset = cryodata(imgs_path, output_path, flag=FLAG, mode = mode, transform=transforms.ToTensor())
    random_seed = 42
    validation_split = .2
    shuffle_dataset = True

    dataset_size = len(dataset)
    indices = list(range(dataset_size))
    split = int(np.floor((1-validation_split) * dataset_size))

    if shuffle_dataset:
        np.random.seed(random_seed)
        np.random.shuffle(indices)
    train_indices, val_indices = indices[:split], indices[split:]

    train_sampler = SubsetRandomSampler(train_indices)
    valid_sampler = SubsetRandomSampler(val_indices)
    #print('the train set size is: {} images'.format(len(train_sampler)))
    #print('the validation set size is: {} images'.format(len(valid_sampler)))
    train_loader = DataLoader(dataset, batch_size=batch_size, sampler=train_sampler)
    validation_loader = DataLoader(dataset, batch_size=batch_size, sampler=valid_sampler)
    sum_, squared_sum_, num_batches = 0, 0, 0
    for img, nm_amplitudes in train_loader:
        sum_ += torch.mean(img, dim=[0, 2, 3])
        squared_sum_ += torch.mean(img**2, dim=[0, 2, 3])
        num_batches += 1
    mean = sum_/num_batches
    std = (squared_sum_/num_batches - mean**2)**0.5
    return mean, std
    
def train(imgs_path, output_path, epochs=400, batch_size=2, lr=1e-4, flag=0, device=0, mode='train'):

    num_epochs = epochs
    random_seed = 42
    validation_split = .5
    shuffle_dataset = True
    FLAG = ''
    if flag==0:
        FLAG = 'nma'
    elif flag==1:
        FLAG = 'ang'
    elif flag==2:
        FLAG = 'shf'
    else:
        FLAG = 'all'
    DEVICE = ''
    if device==0:
        DEVICE = 'cuda'
    else:
        DEVICE = 'cpu'
    mean, std = norm(imgs_path, output_path, FLAG, mode, batch_size)
    transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((mean), (std))])
    transform1 = transforms.Compose([transforms.ToTensor(),
                                    transforms.RandomRotation((-45, 45)),
                                     transforms.Normalize((mean), (std))])
    dataset1 = cryodata(imgs_path, output_path, flag=FLAG, mode= mode,
                        transform=transform)
    dataset2 = cryodata(imgs_path, output_path, flag=FLAG, mode= mode,
                        transform=transform1)
    transform2 = transforms.Compose([transforms.ToTensor(),
                                    transforms.RandomRotation((-90, 90)),
                                     transforms.Normalize((mean), (std))])
    dataset3 = cryodata(imgs_path, output_path, flag=FLAG, mode= mode,
                        transform=transform2)
    increased_dataset = torch.utils.data.ConcatDataset([dataset1, dataset2, dataset3])
    #dataset = cryodata(imgs_path, output_path, flag=FLAG, mode = mode, transform=transform)
    dataset_size = len(increased_dataset)
    indices = list(range(dataset_size))
    split = int(np.floor((1-validation_split) * dataset_size))

    if shuffle_dataset:
        np.random.seed(random_seed)
        np.random.shuffle(indices)
    train_indices, val_indices = indices[:split], indices[split:]

    train_sampler = SubsetRandomSampler(train_indices)
    valid_sampler = SubsetRandomSampler(val_indices)
    print('the train set size is: {} images'.format(len(train_sampler)//3))
    print('the validation set size is: {} images'.format(len(valid_sampler)//3))
    train_loader = DataLoader(increased_dataset, batch_size=batch_size, sampler=train_sampler)
    validation_loader = DataLoader(increased_dataset, batch_size=batch_size, sampler=valid_sampler)
    
    im, p = next(iter(train_loader))
    if FLAG=='nma':
        model = deephemnma(p.shape[1]).to(DEVICE)
    elif FLAG=='ang':
        model = deephemnma(p.shape[1]).to(DEVICE)
    elif FLAG=='shf':
        model = deephemnma(p.shape[1]).to(DEVICE)
    elif FLAG=='all':
        model = deephemnma(p.shape[1]).to(DEVICE)

    optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=1e-5)
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', patience=10)
    criterion = nn.L1Loss()
    writer = SummaryWriter(output_path+'/scalars')
    for epoch in range(num_epochs):

        epoch_loss = 0.0
        running_loss = 0.0

        for img, params in train_loader:
            img = img/255.
            optimizer.zero_grad()
            pred_params = model(img.to(DEVICE), 'train')
            l = criterion(params.to(DEVICE), pred_params)
            l.backward()
            optimizer.step()
            running_loss += l.item()
            epoch_loss += pred_params.shape[0] * l.item()
            print('epoch [{}/{}], loss:{:.4f}'
                  .format(epoch + 1, num_epochs, l.item() ), end='\r')
        valid_loss = 0.0
        with torch.no_grad():
            for img, params in validation_loader:
                pred_params = model(img.to(DEVICE), 'validation')
                l = criterion(params.to(DEVICE), pred_params)
                valid_loss += pred_params.shape[0] * l.item()

        print('epoch [{}/{}], train loss:{:.4f}, validation loss:{:.4f}'
              .format(epoch + 1, num_epochs, epoch_loss / len(train_loader.dataset), valid_loss / len(validation_loader.dataset)))
        writer.add_scalar('Loss/train', epoch_loss / len(train_loader.dataset), epoch+1)
        writer.add_scalar('Loss/validation', valid_loss / len(validation_loader.dataset), epoch+1)
        scheduler.step(epoch_loss)
        torch.save(model.state_dict(), output_path+'/weights.pth')

if __name__ == '__main__':
    train(sys.argv[1],
          sys.argv[2],
          int(sys.argv[3]),
          int(sys.argv[4]),
          float(sys.argv[5]),
          int(sys.argv[6]),
          int(sys.argv[7]))


from pwem.constants import *
from pwem.wizards import *
from continuousflex.protocols.protocol_denoise_volumes import FlexProtVolumeDenoise

class FlexFilterVolumesWizard(FilterVolumesWizard):
    _targets = [(FlexProtVolumeDenoise, ['lowFreqDig','freqDecayDig'])]

    def _getParameters(self, protocol):
        protParams = {}

        labels = ['lowFreqDig', 'highFreqDig', 'freqDecayDig']
        protParams['unit'] = UNIT_PIXEL

        values = [protocol.getAttributeValue(l) for l in labels]

        protParams['input']= protocol.inputVolumes
        protParams['label']= labels
        protParams['value']= values
        protParams['mode'] = 0 # FILTER_SPACE_FOURIER = 0
        return protParams

    def _getProvider(self, protocol):
        _objs = self._getParameters(protocol)['input']
        return FilterVolumesWizard._getListProvider(self, _objs)

    def show(self, form):
        params = self._getParameters(form.protocol)
        _value = params['value']
        _label = params['label']
        _mode = params['mode']
        _unit = params['unit']
        FilterVolumesWizard.show(self, form, _value, _label, _mode, _unit)


# **************************************************************************
# *
# * Author:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# * IMPMC, UPMC, Sorbonne University
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pwem.protocols import ProtImportPdb
from pwem.tests.workflows import TestWorkflow
from pyworkflow.tests import setupTestProject, DataSet
from continuousflex.protocols import (FlexProtNMA, FlexProtSynthesizeSubtomo,NMA_CUTOFF_ABS)
from continuousflex.protocols.protocol_subtomograms_synthesize import MODE_RELATION_MESH
from continuousflex.protocols.protocol_pdb_dimred import FlexProtDimredPdb
from continuousflex.protocols.protocol_subtomograms_classify import FlexProtSubtomoClassify


class TestSubtomogramSynthesize(TestWorkflow):
    """ subtomogram synthesize test """
    
    @classmethod
    def setUpClass(cls):    
        # Create a new project
        setupTestProject(cls)
        cls.ds = DataSet.getDataSet('nma_V2.0')
    
    def test_synthesize_all(self):
        """ Run NMA then synthesize sybtomograms"""
        
        # Import a PDB
        protImportPdb = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                         pdbFile=self.ds.getFile('pdb'))
        protImportPdb.setObjLabel('AK.pdb')
        self.launchProtocol(protImportPdb)
        
        # Launch NMA for PDB imported
        protNMA = self.newProtocol(FlexProtNMA,
                                    cutoffMode=NMA_CUTOFF_ABS)
        protNMA.inputStructure.set(protImportPdb.outputPdb)
        protNMA.setObjLabel('NMA')
        self.launchProtocol(protNMA)

        # Synthesize subtomograms with Mesh relationship
        protSynthesize3 = self.newProtocol(FlexProtSynthesizeSubtomo,
                                         modeList='7-8',
                                         modeRelationChoice=MODE_RELATION_MESH)
        protSynthesize3.inputModes.set(protNMA.outputModes)
        protSynthesize3.setObjLabel('synthesized mesh')
        self.launchProtocol(protSynthesize3)

        protpdbdimred3 = self.newProtocol(FlexProtDimredPdb,
                                         reducedDim=3)
        protpdbdimred3.pdbs.set(protSynthesize3)
        protpdbdimred3.setObjLabel('pdb dimred')
        self.launchProtocol(protpdbdimred3)

        protclassifyhierarchical3= self.newProtocol(FlexProtSubtomoClassify,
                                                   numOfClasses=3)
        protclassifyhierarchical3.ProtSynthesize.set(protSynthesize3)
        protclassifyhierarchical3.setObjLabel('hierarchical')
        self.launchProtocol(protclassifyhierarchical3)
        protclassifyKmeans3 = self.newProtocol(FlexProtSubtomoClassify,
                                                     numOfClasses=3,
                                                     classifyTechnique=1,
                                                     reducedDim=3)
        protclassifyKmeans3.ProtSynthesize.set(protSynthesize3)
        protclassifyKmeans3.setObjLabel('Kmeans')
        self.launchProtocol(protclassifyKmeans3)

        


# * Authors:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# *           Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import os
import time

import pwem.emlib.metadata as md
import pyworkflow.protocol.params as params
from pwem.protocols import ProtAnalysis3D
from pyworkflow.protocol.params import NumericRangeParam
from pyworkflow.utils import getListFromRangeString
from pyworkflow.utils import replaceExt
import numpy as np
from math import cos, sin, pi
from pwem.objects import AtomStruct

from continuousflex.protocols.utilities.pdb_handler import ContinuousFlexPDBHandler

MODE_RELATION_LINEAR = 0
MODE_RELATION_3CLUSTERS = 1
MODE_RELATION_MESH = 2
MODE_RELATION_RANDOM = 3
MODE_RELATION_PARABOLA = 4


class FlexProtSynthesizePDBs(ProtAnalysis3D):
    """ Protocol for synthesizing flexible PDBs using Normal Mode Analysis. """
    _label = 'synthesize PDBs'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection('Input')
        form.addParam('inputModes', params.PointerParam,
                      pointerClass='SetOfNormalModes',
                      label="Normal modes",
                      help='Set of modes computed by normal mode analysis.')
        form.addParam('modeList', NumericRangeParam,
                      label="Modes selection",
                      allowsNull=True,
                      default='7-8',
                      help='Select the normal modes that will be used for PDB synthesis.'
                           'It is usually two modes that should be selected, unless if the relationship is linear or '
                           'random.\n '
                           'You have several ways to specify the modes.\n'
                           'Examples:\n'
                           ' "7,8-10" -> [7,8,9,10]\n'
                           ' "8, 10, 12" -> [8,10,12]\n'
                           ' "8 9, 10-12" -> [8,9,10,11,12])\n')
        form.addParam('modeRelationChoice', params.EnumParam, default=MODE_RELATION_LINEAR,
                      choices=['Linear relationship', '3 clusters', 'Grid', 'Random', 'Parabolic'],
                      label='Relationship between the modes',
                      help='linear relationship: all the selected modes will have equal amplitudes. \n'
                           '3 clusters: the volumes will be divided exactly into three classes.\n'
                           'Grid: the amplitudes will be in a grid shape (grid size is square of what grid step).\n'
                           'Random: all the amplitudes will be random in the given range.\n'
                           'Parabolic: The relationship between two modes will resembles an upper half circle.')
        form.addParam('centerPoint', params.IntParam, default=100,
                      condition='modeRelationChoice==%d' % MODE_RELATION_3CLUSTERS,
                      label='Center point',
                      help='This number will be used to determine the distance between the clusters'
                           'center1 = (-center_point, 0)'
                           'center2 = (center_point, 0)'
                           'center3 = (0, center_point)')
        form.addParam('modesAmplitudeRange', params.IntParam, default=150,
                      allowsNull=True,
                      condition='modeRelationChoice != %d' % MODE_RELATION_3CLUSTERS,
                      label='Amplitude range N --> [-N, N]',
                      help='Choose the number N for which the generated normal mode amplitudes are in the range of'
                           ' [-N, N]')
        form.addParam('meshRowPoints', params.IntParam, default=6,
                      allowsNull=True,
                      condition='modeRelationChoice==%d' % MODE_RELATION_MESH,
                      label='Grid number of steps',
                      help='This number will be the number of points in the row and the column (grid shape will be '
                           'size*size)')
        form.addParam('numberOfPDBs', params.IntParam, default=36,
                      label='Number of PDBs',
                      condition='modeRelationChoice!=%d' % MODE_RELATION_MESH,
                      help='Number of atomic structures that will be generated')
        form.addParam('seedOption', params.BooleanParam, default=True,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='Random seed',
                      help='Keeping it as True means that different runs will generate different PDBs in terms '
                           'of conformations. If you set as False, then different runs will '
                           'have the same conformations and angles '
                           '(setting to False allows you to generate the same conformations and orientations with '
                           'different noise values).')

        # form.addParallelSection(threads=0, mpi=8)
        # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        if self.seedOption.get():
            np.random.seed(int(time.time()))
        else:
            np.random.seed(0)
        self._insertFunctionStep("generateDeformationsStep")
        self._insertFunctionStep("createOutputStep")

    # --------------------------- STEPS functions --------------------------------------------
    def generateDeformationsStep(self):
        # Find the right PDB file to use for data synthesis
        pdb_name1 = os.path.dirname(self.inputModes.get().getFileName()) + '/atoms.pdb'
        pdb_name2 = os.path.dirname(self.inputModes.get().getFileName()) + '/pseudoatoms.pdb'
        if os.path.exists(pdb_name1):
            fnPDB = pdb_name1
        else:
            fnPDB = pdb_name2
        # use the input relationship between the modes to generate normal mode amplitudes metadata
        fnModeList = replaceExt(self.inputModes.get().getFileName(), 'xmd')
        modeAmplitude = self.modesAmplitudeRange.get()
        meshRowPoints = self.meshRowPoints.get()
        numberOfModes = self.inputModes.get().getSize()
        modeSelection = np.array(getListFromRangeString(self.modeList.get()))
        deformationFile = self._getExtraPath('GroundTruth.xmd')
        pdbMD = md.MetaData()
        # these vairables for alternating the generation when 3 clusters are selected
        cluster1 = cluster2 = cluster3 = False
        # these variables for the generaton of a mesh if selected
        XX, YY = np.meshgrid(np.linspace(start=-modeAmplitude, stop=modeAmplitude, num=meshRowPoints),
                             np.linspace(start=modeAmplitude, stop=-modeAmplitude, num=meshRowPoints))
        mode7_samples = XX.reshape(-1)
        mode8_samples = YY.reshape(-1)

        # iterate over the number of outputs (if mesh, this has to be calculated)
        numberOfPDBs = self.getNumberOfPdbs()

        for i in range(numberOfPDBs):
            deformations = np.zeros(numberOfModes)

            if self.modeRelationChoice == MODE_RELATION_LINEAR:
                amplitude = self.modesAmplitudeRange.get()
                deformations[modeSelection - 1] = np.ones(len(modeSelection)) * np.random.uniform(-amplitude, amplitude)
            elif self.modeRelationChoice == MODE_RELATION_3CLUSTERS:
                center_point = self.centerPoint.get()
                center1 = (-center_point, 0)
                center2 = (center_point, 0)
                center3 = (0, center_point)
                if not (cluster1 or cluster2 or cluster3):
                    cluster1 = True
                if cluster3:
                    deformations[modeSelection - 1] = center3
                    cluster3 = False
                if cluster2:
                    deformations[modeSelection - 1] = center2
                    cluster2 = False
                    cluster3 = True
                if cluster1:
                    deformations[modeSelection - 1] = center1
                    cluster1 = False
                    cluster2 = True
            elif self.modeRelationChoice == MODE_RELATION_RANDOM:
                amplitude = self.modesAmplitudeRange.get()
                deformations[modeSelection - 1] = np.random.uniform(-amplitude, amplitude, len(modeSelection))
            elif self.modeRelationChoice == MODE_RELATION_PARABOLA:
                amplitude = self.modesAmplitudeRange.get()
                rv = np.random.uniform(0, 1)
                point = (amplitude * cos(rv * pi), amplitude * sin(rv * pi))
                deformations[modeSelection - 1] = point

            elif self.modeRelationChoice == MODE_RELATION_MESH:
                new_point = (mode7_samples[i], mode8_samples[i])
                deformations[modeSelection - 1] = new_point

            # we won't keep the first 6 modes
            deformations = deformations[6:]

            self.nma_deform_pdb(fnPDB, fnModeList, self._getExtraPath(str(i + 1).zfill(5) + '_df.pdb'), deformations)

            pdbMD.setValue(md.MDL_IMAGE, self._getExtraPath(str(i + 1).zfill(5) + '_df.pdb'),
                           pdbMD.addObject())
            pdbMD.setValue(md.MDL_NMA, list(deformations), i + 1)

        pdbMD.write(deformationFile)

    def nma_deform_pdb(self, fnPDB, fnModeList, fnOut, deformList):

        def readModes(fnIn):
            modesMD = md.MetaData(fnIn)
            vectors = []
            for objId in modesMD:
                vecFn = modesMD.getValue(md.MDL_NMA_MODEFILE, objId)
                vec = np.loadtxt(vecFn)
                vectors.append(vec)
            return vectors

        pdb = ContinuousFlexPDBHandler(fnPDB)
        modes = readModes(fnModeList)
        for i in range(len(deformList)):
            pdb.coords += deformList[i] * modes[7 - 1 + i]
        pdb.write_pdb(fnOut)

    def createOutputStep(self):
        pdbset = self._createSetOfPDBs("outputPDBs")
        pdbMD = md.MetaData(self._getExtraPath('GroundTruth.xmd'))
        for objID in pdbMD:
            filename = pdbMD.getValue(md.MDL_IMAGE, objID)
            pdb = AtomStruct(filename=filename)
            pdbset.append(pdb)

        self._defineOutputs(outputPDBs=pdbset)

    def getNumberOfPdbs(self):
        if self.modeRelationChoice.get() is MODE_RELATION_MESH:
            numberOfPDBs = self.meshRowPoints.get() ** 2
        else:
            numberOfPDBs = self.numberOfPDBs.get()
        return numberOfPDBs

    def getInputPdb(self):
        """ Return the Pdb object associated with the normal modes. """
        return self.inputModes.get().getPdb()

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd', 
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def _getLocalModesFn(self):
        modesFn = self.inputModes.get().getFileName()
        return self._getBasePath(modesFn)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Mohamad Harastani (mohamad.harastani@igbmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

"""

@article{Jin2014,
title = "Iterative Elastic 3D-to-2D Alignment Method Using Normal Modes for Studying Structural Dynamics of Large Macromolecular Complexes",
journal = "Structure",
volume = "22",
pages = "1 - 11",
year = "2014",
doi = "http://dx.doi.org/10.1016/j.str.2014.01.004",
url = "http://www.ncbi.nlm.nih.gov/pubmed/24508340",
author = "Jin, Q. and Sorzano, C. O. S. and de la Rosa-Trevín, J. M. and Bilbao-Castro, J. R. and Núñez-Ramirez, R. and Llorca, O. and Tama, F. and Jonic, S.",
keywords = "Normal mode analysis, NMA "
}

@article{Jonic2005,
title = "Spline-based image-to-volume registration for three-dimensional electron microscopy ",
journal = "Ultramicroscopy ",
volume = "103",
number = "4",
pages = "303 - 317",
year = "2005",
issn = "0304-3991",
doi = "http://dx.doi.org/10.1016/j.ultramic.2005.02.002",
url = "http://www.sciencedirect.com/science/article/pii/S0304399105000173",
author = "Jonic, S. and C.O.S. Sorzano and P. Thevenaz and C. El-Bez and S. De Carlo and M. Unser",
keywords = "2D/3D registration, Splines, 3DEM, Angular assignment "
}

@article{Nogales2013,
title={3DEM Loupe: analysis of macromolecular dynamics using structures from electron microscopy},
author={Nogales-Cadenas, R. and Jonic, S. and Tama, F. and Arteni, A. A. and Tabas-Madrid, D. and V{\'a}zquez, M. and Pascual-Montano, A. and Sorzano, C. O. S.},
journal={Nucleic acids research},
year={2013},
publisher={Oxford Univ Press},
doi={http://dx.doi.org/10.1093/nar/gkt385}
}


@article{Sorzano2004b,
    volume = "146", 
    doi = "http://dx.doi.org/10.1016/j.jsb.2004.01.006", 
    author = "Sorzano, C.O.S. and S. Jonic and C. El-Bez and J.M. Carazo and S. De Carlo and P. Thevenaz and M. Unser", 
    title = "A multiresolution approach to orientation assignment in 3D electron microscopy of single particles ", 
    journal = "JSB ", 
    issn = "1047-8477", 
    number = "3", 
    note = "", 
    link = "http://www.sciencedirect.com/science/article/pii/S1047847704000073", 
    year = "2004", 
    pages = "381 - 392"
}

@article{Sorzano2016,
title = "StructMap: Elastic distance analysis of electron microscopy maps for studying conformational changes",
journal = "Biophysical J.",
volume = "110",
number = "",
pages = "1753-1765",
year = "2016",
note = "",
issn = "",
doi = "http://doi.org/10.1016/j.bpj.2016.03.019",
url = "http://doi.org/10.1016/j.bpj.2016.03.019",
author = "C.O.S. Sorzano, A.L. Álvarez-Cabrera, M. Kazemi, J.M. Carazo, S. Jonic",
keywords = ""
}


@article{harastani2020hybrid,
  title={Hybrid Electron Microscopy Normal Mode Analysis with Scipion},
  author={Harastani, Mohamad and Sorzano, Carlos Oscar S and Joni{\'c}, Slavica},
  journal={Protein Science},
  volume={29},
  number={1},
  pages={223--236},
  year={2020},
  publisher={Wiley Online Library},
  doi= {https://doi.org/10.1002/pro.3772}
}

@article{moebel2020monte,
  title={A Monte Carlo framework for missing wedge restoration and noise removal in cryo-electron tomography},
  author={Moebel, Emmanuel and Kervrann, Charles},
  journal={Journal of Structural Biology: X},
  volume={4},
  pages={100013},
  year={2020},
  publisher={Elsevier}
}

@article{vuillemot2022NMMD,
title = {NMMD: Efficient Cryo-EM Flexible Fitting Based on Simultaneous Normal Mode and Molecular Dynamics atomic displacements},
journal = {Journal of Molecular Biology},
volume = {434},
number = {7},
pages = {167483},
year = {2022},
issn = {0022-2836},
doi = {https://doi.org/10.1016/j.jmb.2022.167483},
url = {https://www.sciencedirect.com/science/article/pii/S0022283622000523},
author = {Rémi Vuillemot and Osamu Miyashita and Florence Tama and Isabelle Rouiller and Slavica Jonic}
}

@article{kobayashi2017genesis,
author = {Kobayashi, Chigusa and Jung, Jaewoon and Matsunaga, Yasuhiro and Mori, Takaharu and Ando, Tadashi and Tamura, Koichi and Kamiya, Motoshi and Sugita, Yuji},
title = {GENESIS 1.1: A hybrid-parallel molecular dynamics simulator with enhanced sampling algorithms on multiple computational platforms},
journal = {Journal of Computational Chemistry},
volume = {38},
number = {25},
pages = {2193-2206},
keywords = {molecular dynamics, string method, replica exchange molecular dynamics, graphics processing unit, multiple time step integration},
doi = {https://doi.org/10.1002/jcc.24874},
url = {https://onlinelibrary.wiley.com/doi/abs/10.1002/jcc.24874},
eprint = {https://onlinelibrary.wiley.com/doi/pdf/10.1002/jcc.24874},
abstract = {GENeralized-Ensemble SImulation System (GENESIS) is a software package for molecular dynamics (MD) simulation of biological systems. It is designed to extend limitations in system size and accessible time scale by adopting highly parallelized schemes and enhanced conformational sampling algorithms. In this new version, GENESIS 1.1, new functions and advanced algorithms have been added. The all-atom and coarse-grained potential energy functions used in AMBER and GROMACS packages now become available in addition to CHARMM energy functions. The performance of MD simulations has been greatly improved by further optimization, multiple time-step integration, and hybrid (CPU + GPU) computing. The string method and replica-exchange umbrella sampling with flexible collective variable choice are used for finding the minimum free-energy pathway and obtaining free-energy profiles for conformational changes of a macromolecule. These new features increase the usefulness and power of GENESIS for modeling and simulation in biological research. © 2017 Wiley Periodicals, Inc.},
year = {2017}
}

@article{CHEN2013235,
title = {Fast and accurate reference-free alignment of subtomograms},
journal = {Journal of Structural Biology},
volume = {182},
number = {3},
pages = {235-245},
year = {2013},
issn = {1047-8477},
doi = {https://doi.org/10.1016/j.jsb.2013.03.002},
url = {https://www.sciencedirect.com/science/article/pii/S1047847713000737},
author = {Yuxiang Chen and Stefan Pfeffer and Thomas Hrabe and Jan Michael Schuller and Friedrich Förster},
keywords = {Cryo-electron tomography, Subtomogram averaging, Spherical harmonics},
abstract = {In cryoelectron tomography alignment and averaging of subtomograms, each dnepicting the same macromolecule, improves the resolution compared to the individual subtomogram. Major challenges of subtomogram alignment are noise enhancement due to overfitting, the bias of an initial reference in the iterative alignment process, and the computational cost of processing increasingly large amounts of data. Here, we propose an efficient and accurate alignment algorithm via a generalized convolution theorem, which allows computation of a constrained correlation function using spherical harmonics. This formulation increases computational speed of rotational matching dramatically compared to rotation search in Cartesian space without sacrificing accuracy in contrast to other spherical harmonic based approaches. Using this sampling method, a reference-free alignment procedure is proposed to tackle reference bias and overfitting, which also includes contrast transfer function correction by Wiener filtering. Application of the method to simulated data allowed us to obtain resolutions near the ground truth. For two experimental datasets, ribosomes from yeast lysate and purified 20S proteasomes, we achieved reconstructions of approximately 20Å and 16Å, respectively. The software is ready-to-use and made public to the community.}
}

@article{harastani2022continuousflex,
  title={ContinuousFlex: Software package for analyzing continuous conformational variability of macromolecules in cryo electron microscopy and tomography data},
  author={Harastani, Mohamad and Vuillemot, R{\'e}mi and Hamitouche, Ilyes and Moghadam, Nima Barati and Jonic, Slavica},
  journal={Journal of Structural Biology},
  volume={214},
  number={4},
  pages={107906},
  year={2022},
  publisher={Elsevier}
}

@article{harastani2022tomoflow,
  title={TomoFlow: Analysis of continuous conformational variability of macromolecules in cryogenic subtomograms based on 3D dense optical flow},
  author={Harastani, Mohamad and Eltsov, Mikhail and Leforestier, Am{\'e}lie and Jonic, Slavica},
  journal={Journal of molecular biology},
  volume={434},
  number={2},
  pages={167381},
  year={2022},
  publisher={Elsevier}
}

@article{harastani2021hemnma,
  title={Hemnma-3d: Cryo electron tomography method based on normal mode analysis to study continuous conformational variability of macromolecular complexes},
  author={Harastani, Mohamad and Eltsov, Mikhail and Leforestier, Am{\'e}lie and Jonic, Slavica},
  journal={Frontiers in molecular biosciences},
  pages={317},
  year={2021},
  publisher={Frontiers}
}

@article{hamitouche2022deephemnma,
  title={DeepHEMNMA: ResNet-based hybrid analysis of continuous conformational heterogeneity in cryo-EM single particle images},
  author={Hamitouche, Ilyes and Jonic, Slavica},
  journal={Frontiers in Molecular Biosciences},
  year={2022}
}

"""




# **************************************************************************
# * Author:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pyworkflow.object import String
from pyworkflow.protocol.params import (PointerParam, EnumParam, IntParam)
from pwem.protocols import ProtAnalysis3D
from pyworkflow.utils.path import makePath, copyFile, removeBaseExt
from pyworkflow.protocol import params
from sklearn.cluster import AgglomerativeClustering, KMeans
import time
import os
from sh_alignment.tompy.transform import fft, ifft, fftshift, ifftshift
import pwem.emlib.metadata as md
from continuousflex.protocols.utilities.spider_files3 import save_volume, open_volume
import xmipp3
from pwem.objects import Volume
import numpy as np
from sklearn import decomposition
from joblib import dump, load
from pwem.utils import runProgram


class FlexProtSubtomoClassify(ProtAnalysis3D):
    """ Protocol applying post alignment classification on subtomograms. """
    _label = 'classify subtomograms'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('SubtomoSource', EnumParam, default=0,
                      label='Source of Subtomograms',
                      choices=['After subtomogram synthesis', 'After subtomogram averaging'],
                      help='Choose the source of the subtomograms that you want to classify')
        form.addParam('ProtSynthesize', params.PointerParam, pointerClass='FlexProtSynthesizeSubtomo',
                      condition='SubtomoSource == 0',
                      label="Subtomogram synthesis",
                      help='All PDBs should have the same size')
        form.addParam('StA',params.PointerParam,
                      condition='SubtomoSource == 1',
                      pointerClass='FlexProtSubtomogramAveraging',
                      label="StA protocol",
                      help='Choose a subtomogram averaging previous run')
        form.addParam('classifyTechnique', EnumParam, default=0,
                      label='Classification techinque',
                      choices=['Hierarchical clustering', 'Dimentionality reduction then Clustering'],
                      help='Choose a classification techinque')
        form.addParam('ClusteringLinkage', EnumParam, default=0,
                      label='Linkage',
                      condition='classifyTechnique == 0',
                      choices=['ward'])
        form.addParam('dimredMethod', EnumParam, default=0,
                      condition='classifyTechnique == 1',
                      choices=['Scikit-Learn PCA'],
                      label='Dimensionality reduction method',
                      help='This method will be used to reduce the dimensions of the covariance matrix')
        form.addParam('applyMask', params.BooleanParam, label='Use a mask?', default=False,
                       help='This mask will be applied on the aligned particles before finding the cross correlation'
                            ', a proper mask is a mask derived from the subtomogram average (tight), that can be '
                            'applied on it without cropping any part.')
        form.addParam('Mask', params.PointerParam,
                       condition='applyMask',
                       pointerClass='Volume', allowsNull=True,
                       label="Select mask")
        form.addParam('reducedDim', IntParam, default=2,
                      condition='classifyTechnique == 1',
                      label='Reduced dimension')
        form.addParam('numOfClasses', IntParam, default=2,
                      label='Number of classes')
        # form.addParallelSection(threads=0, mpi=8)

        # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        mdSubtomo = self.getSubtomoMetaData()
        self._insertFunctionStep('subtomo_wedge_align',mdSubtomo)
        self._insertFunctionStep('find_covariance_matrix')
        if self.classifyTechnique.get() == 0:
            self._insertFunctionStep('performHierarchicalClustering')
        else:
            self._insertFunctionStep('performKmeansClustering')
        self._insertFunctionStep('findTotalAverage')
        self._insertFunctionStep('createOutputStep')


    # --------------------------- STEPS functions --------------------------------------------
    def subtomo_wedge_align(self,mdSubtomo):
        # we align the subtomograms but we align also a missing wedge mask for each one
        tilt = self.getTiltRange()
        # Creating a missing wedge mask:
        start_ang = tilt[0]
        end_ang = tilt[1]
        leg = self.getVolumeSize()
        size = (leg, leg, leg)
        MW_mask = np.ones(size)
        x, z = np.mgrid[0.:size[0], 0.:size[2]]
        x -= size[0] / 2
        ind = np.where(x)
        z -= size[2] / 2
        angles = np.zeros(z.shape)
        angles[ind] = np.arctan(z[ind] / x[ind]) * 180 / np.pi
        angles = np.reshape(angles, (size[0], 1, size[2]))
        angles = np.repeat(angles, size[1], axis=1)
        MW_mask[angles > -start_ang] = 0
        MW_mask[angles < -end_ang] = 0
        MW_mask[size[0] // 2, :, :] = 0
        MW_mask[size[0] // 2, :, size[2] // 2] = 1
        fnmask = self._getExtraPath('missing_wedge.spi')
        save_volume(np.float32(MW_mask), fnmask)
        args = ' -i ' + fnmask + ' --rotate_volume euler 0 90 0'
        runProgram('xmipp_transform_geometry',args)
        # Now aligning each subtomogram and its missing wedge version
        mw_path = self._getExtraPath('mw_masks/')
        subtom_aligned_path = self._getExtraPath('aligned_subtomograms/')
        makePath(mw_path)
        makePath(subtom_aligned_path)
        subtomogramMD = md.MetaData(mdSubtomo)
        subtomogaligneMD = md.MetaData()
        mwalignedMD = md.MetaData()
        for i in subtomogramMD:
            fnsubtomo = subtomogramMD.getValue(md.MDL_IMAGE, i)
            bnsubtomo = os.path.basename(fnsubtomo)
            bnwedge = removeBaseExt(bnsubtomo)+'_wedge.spi'
            fnalignedsubtomo = self._getExtraPath('aligned_subtomograms/'+bnsubtomo)
            fnalignedmask = self._getExtraPath('mw_masks/'+bnwedge)
            # print(fnalignedsubtomo)
            # print(fnalignedmask)
            rot = str(subtomogramMD.getValue(md.MDL_ANGLE_ROT, i))
            tilt = str(subtomogramMD.getValue(md.MDL_ANGLE_TILT, i))
            psi = str(subtomogramMD.getValue(md.MDL_ANGLE_PSI, i))
            shiftx = str(subtomogramMD.getValue(md.MDL_SHIFT_X, i))
            shifty = str(subtomogramMD.getValue(md.MDL_SHIFT_Y, i))
            shiftz = str(subtomogramMD.getValue(md.MDL_SHIFT_Z, i))
            # align the subtomogram

            if self.getAngleY() == 90:
                args = '-i ' + fnsubtomo + ' -o ' + fnalignedsubtomo + ' --rotate_volume euler 0 90 0'
                runProgram('xmipp_transform_geometry', args)
                params = '-i ' + fnalignedsubtomo + ' -o ' + fnalignedsubtomo + ' '

            else:
                params = '-i ' + fnsubtomo + ' -o ' + fnalignedsubtomo + ' '
                params += ' --inverse '

            params += '--rotate_volume euler ' + rot + ' ' + tilt + ' ' + psi + ' '
            params += '--shift ' + shiftx + ' ' + shifty + ' ' + shiftz + ' '

            runProgram('xmipp_transform_geometry', params)

            # align the mask (no shift should be applied only angles)
            if self.getAngleY() == 90:
                args = '-i ' + fnmask + ' -o ' + fnalignedmask + ' --rotate_volume euler 0 90 0'
                runProgram('xmipp_transform_geometry', args)
                params = '-i ' + fnalignedmask + ' -o ' + fnalignedmask + ' '
            else:
                params = '-i ' + fnmask + ' -o ' + fnalignedmask + ' '
                params += ' --inverse '

            params += '--rotate_volume euler ' + rot + ' ' + tilt + ' ' + psi + ' '

            runProgram('xmipp_transform_geometry', params)

            if (self.applyMask):
                maskfn = self.Mask.get().getFileName()
                params = '-i ' + fnalignedsubtomo + ' -o ' + fnalignedsubtomo + ' --mult ' + maskfn
                runProgram('xmipp_image_operate', params)

            subtomogaligneMD.setValue(md.MDL_IMAGE, fnalignedsubtomo, subtomogaligneMD.addObject())
            mwalignedMD.setValue(md.MDL_IMAGE, fnalignedmask, mwalignedMD.addObject())
        subtomogaligneMD.write(self._getExtraPath('aligned_subtomograms.xmd'))
        mwalignedMD.write(self._getExtraPath('aligned_masks.xmd'))




    def find_covariance_matrix(self):
        # if the covariance matrix exists, we should not re-evaluate it (it takes long time)
        fn_covarmat = self._getExtraPath('covar_mat.pkl')
        if os.path.exists(fn_covarmat):
            pass
        subtomogaligneMD= md.MetaData(self._getExtraPath('aligned_subtomograms.xmd'))
        mwalignedMD= md.MetaData(self._getExtraPath('aligned_masks.xmd'))
        N = subtomogaligneMD.size()
        X = np.zeros([N, N])
        print(N)
        A = time.time()
        for i in subtomogaligneMD:
            if i == 2:
                B = time.time()-A
                print('estimated time to finish is ', B*N/2)
            name_i = subtomogaligneMD.getValue(md.MDL_IMAGE, i)
            wedge_i = mwalignedMD.getValue(md.MDL_IMAGE, i)
            Vi = open_volume(name_i)
            Wi = open_volume(wedge_i)
            print('line is', str(i),' out of ',str(N))
            for j in range(i,N+1):
                name_j = subtomogaligneMD.getValue(md.MDL_IMAGE, j)
                wedge_j = mwalignedMD.getValue(md.MDL_IMAGE, j)
                Vj = open_volume(name_j)
                Wj = open_volume(wedge_j)
                Omega = Wi * Wj
                FVi = fft(Vi)
                FVi = fftshift(FVi) * Omega
                FVi = ifftshift(FVi)
                Vi_p = ifft(FVi)
                FVj = fft(Vj)
                FVj = fftshift(FVj) * Omega
                FVj = ifftshift(FVj)
                Vj_p = ifft(FVj)
                Vi_p = np.array(Vi_p, dtype=np.float32)
                Vj_p = np.array(Vj_p, dtype=np.float32)
                X[i-1, j-1] = self.cc(Vi_p, Vj_p)
                X[j-1, i-1] = self.cc(Vj_p, Vi_p)
        # save the covariance matrix:
        dump(X, fn_covarmat)

    def performHierarchicalClustering(self):
        fn_covarmat = self._getExtraPath('covar_mat.pkl')
        data = load(fn_covarmat)
        # 1 - CCCij (to keep with the literature)
        data = np.ones_like(data) - data
        clustering = AgglomerativeClustering(n_clusters=self.numOfClasses.get(), linkage='ward')
        clustering_class = clustering.fit_predict(data)
        labels = clustering.labels_
        # for l in np.unique(label):
        #     print(len(data[label == l]))
        dump(labels, filename=self._getExtraPath('hierarchical_clustering_labels.pkl'))
        print(clustering_class)
        # creating a metadate of subtomograms for each class and an average
        N = self.getVolumeSize()
        Averages = np.zeros([N, N, N, self.numOfClasses.get()])
        subtomogaligneMD = md.MetaData(self._getExtraPath('aligned_subtomograms.xmd'))
        # creating a metadata for each class
        classesMD = [md.MetaData() for i in range(self.numOfClasses.get())]
        for i in subtomogaligneMD:
            name = subtomogaligneMD.getValue(md.MDL_IMAGE, i)
            # the metadata index start from 1, we should subtract one to match the matrix
            label = labels[i-1]
            vol = open_volume(name)
            Averages[:, :, :, label] += vol
            classesMD[label].setValue(md.MDL_IMAGE, name, classesMD[label].addObject())
        for i in range(self.numOfClasses.get()):
            name = self._getExtraPath('class_'+str(i).zfill(2)+'.xmd')
            classesMD[i].write(name)
        # creating a metadata for all class averages:
        md_averages = md.MetaData()
        makePath(self._getExtraPath('class_averages/'))
        for j in range(self.numOfClasses.get()):
            num = len(labels[labels == j])
            print('Number of subtomograms in class #', j, 'is ', num)
            name = self._getExtraPath('class_averages/') + 'cluster' + str(j).zfill(2) + '.spi'
            ave = np.array(Averages[:, :, :, j] / num, dtype=np.float32)
            save_volume(ave, name)
            md_averages.setValue(md.MDL_IMAGE,name,md_averages.addObject())
        md_averages.write(self._getExtraPath('averages.xmd'))


    def performKmeansClustering(self):
        X = load(self._getExtraPath('covar_mat.pkl'))
        pca = decomposition.PCA(n_components=self.reducedDim.get())
        pca.fit(X)
        data = pca.transform(X)
        pca_pickled = self._getExtraPath('pca_pickled.pkl')
        np.savetxt(self._getExtraPath('dimred_mat.txt'),data)
        dump(pca,pca_pickled)
        # clustering now
        clustering = KMeans(n_clusters=self.numOfClasses.get()).fit(data)
        clustering_class = clustering.fit_predict(data)
        labels = clustering.labels_
        # for l in np.unique(label):
        #     print(len(data[label == l]))
        dump(clustering, filename=self._getExtraPath('kmeans_algo.pkl'))
        dump(labels, filename=self._getExtraPath('kmeans_clustering_labels.pkl'))
        print(clustering_class)
        # creating a metadate of subtomograms for each class and an average
        N = self.getVolumeSize()
        Averages = np.zeros([N, N, N, self.numOfClasses.get()])
        subtomogaligneMD = md.MetaData(self._getExtraPath('aligned_subtomograms.xmd'))
        # creating a metadata for each class
        classesMD = [md.MetaData() for i in range(self.numOfClasses.get())]
        for i in subtomogaligneMD:
            name = subtomogaligneMD.getValue(md.MDL_IMAGE, i)
            # the metadata index start from 1, we should subtract one to match the matrix
            label = labels[i - 1]
            vol = open_volume(name)
            Averages[:, :, :, label] += vol
            classesMD[label].setValue(md.MDL_IMAGE, name, classesMD[label].addObject())
        for i in range(self.numOfClasses.get()):
            name = self._getExtraPath('class_' + str(i).zfill(2) + '.xmd')
            classesMD[i].write(name)
        # creating a metadata for all class averages:
        md_averages = md.MetaData()
        makePath(self._getExtraPath('class_averages/'))
        for j in range(self.numOfClasses.get()):
            num = len(labels[labels == j])
            print('Number of subtomograms in class #', j, 'is ', num)
            name = self._getExtraPath('class_averages/') + 'cluster' + str(j).zfill(2) + '.spi'
            ave = np.array(Averages[:, :, :, j] / num, dtype=np.float32)
            save_volume(ave, name)
            md_averages.setValue(md.MDL_IMAGE, name, md_averages.addObject())
        md_averages.write(self._getExtraPath('averages.xmd'))
        pass

    def findTotalAverage(self):
        N = self.getVolumeSize()
        Average = np.zeros([N, N, N])
        subtomogaligneMD = md.MetaData(self._getExtraPath('aligned_subtomograms.xmd'))
        count = 0
        for i in subtomogaligneMD:
            name = subtomogaligneMD.getValue(md.MDL_IMAGE, i)
            vol = open_volume(name)
            Average[:, :, :] += vol
            count+=1
        name = self._getExtraPath('global_average.spi')
        save_volume(np.array(Average[:, :, :] / count, dtype=np.float32), name)
        pass

    def createOutputStep(self):
        out_mdfn = self._getExtraPath('averages.xmd')
        partSet = self._createSetOfVolumes('Averages')
        xmipp3.convert.readSetOfVolumes(out_mdfn, partSet)
        partSet.setSamplingRate(self.getSamplingRate())
        outvolume = Volume()
        outvolume.setSamplingRate(self.getSamplingRate())
        outvolume.setFileName(self._getExtraPath('global_average.spi'))
        self._defineOutputs(ClassAvarages=partSet, GlobalAverage=outvolume)
        pass

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd',
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def getSubtomoMetaData(self):
        if self.SubtomoSource.get()==0:
            return self.ProtSynthesize.get()._getExtraPath('GroundTruth.xmd')
        if self.SubtomoSource.get()==1:
            return self.StA.get()._getExtraPath('final_md.xmd')

    def getTiltRange(self):
        if self.SubtomoSource.get()==0:
            return [self.ProtSynthesize.get().tiltLow.get(), self.ProtSynthesize.get().tiltHigh.get()]
        if self.SubtomoSource.get()==1:
            return [self.StA.get().tiltLow.get(), self.StA.get().tiltHigh.get()]

    def getAngleY(self):
        if self.SubtomoSource.get() == 0:
            return 0
        if self.SubtomoSource.get() == 1:
            temp = md.MetaData(self.StA.get()._getExtraPath('final_md.xmd'))
            return temp.getValue(md.MDL_ANGLE_Y, 1)


    def getSamplingRate(self):
        if self.SubtomoSource.get()==0:
            return self.ProtSynthesize.get().samplingRate.get()
        if self.SubtomoSource.get()==1:
            return self.StA.get().inputVolumes.get().getSamplingRate()

    def getVolumeSize(self):
        if self.SubtomoSource.get() == 0:
            return self.ProtSynthesize.get().volumeSize.get()
        if self.SubtomoSource.get() == 1:
            return self.StA.get().SubtomogramAverage.getDim()[0]


    def getOutputMatrixFile(self):
        return self._getExtraPath('output_matrix.txt')

    def getDeformationFile(self):
        return self._getExtraPath('pdbs_mat.txt')

    def normalize(self,v):
        """Normalize a volume.
        @param v: input volume.
        @return: Normalized volume.
        """
        m = np.mean(v)
        v = v - m
        s = np.std(v)
        v = v / s
        return v

    def cc(self, v1, v2):
        """Compute the Normalized Cross Correlation between the two volumes.
        @param v1: volume 1.
        @param v2: volume 2.
        @return: NCC.
        """
        vv1 = self.normalize(v1)
        vv2 = self.normalize(v2)
        score = np.sum(vv1 * vv2) / vv1.size
        return score



# **************************************************************************
# *
# * Authors:  Carlos Oscar Sanchez Sorzano (coss@cnb.csic.es), May 2013
# *           Slavica Jonic                (slavica.jonic@upmc.fr)
# * Ported to Scipion:
# *           J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es), Jan 2014
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import math
from os.path import basename, exists, join

from pwem.convert.atom_struct import cifToPdb
from pwem.emlib import MetaData, MDL_NMA_ATOMSHIFT, MDL_NMA_MODEFILE
from pyworkflow.utils import redStr, replaceBaseExt
from pyworkflow.utils.path import copyFile, createLink, makePath, cleanPath, moveFile
from pyworkflow.protocol.params import (PointerParam, IntParam, FloatParam,
                                        LEVEL_ADVANCED)
from pwem.objects import SetOfNormalModes
from xmipp3.base import XmippMdRow
from .protocol_nma_base import FlexProtNMABase, NMA_CUTOFF_REL
from .convert import rowToMode, getNMAEnviron
from pwem import Domain


class FlexProtNMA(FlexProtNMABase):
    """ Flexible angular alignment using normal modes """
    _label = 'nma analysis'

    def _defineParams(self, form):
        form.addSection(label='Normal Mode Analysis')
        form.addParam('inputStructure', PointerParam, label="Input structure",
                      important=True,
                      pointerClass='AtomStruct',
                      help='The input structure can be an atomic model '
                           '(true PDB) or a pseudoatomic model\n'
                           '(an EM volume converted into pseudoatoms)')
        FlexProtNMABase._defineParamsCommon(self, form)
        form.addParam('rtbBlockSize', IntParam, default=10,
                      expertLevel=LEVEL_ADVANCED,
                      label='Number of residues per RTB block (for atomic structures)',
                      help='Used only with atoms. Normal modes of atomic structures are computed with the RTB method. '
                           '\n '
                           'This is the RTB block size. In the RTB method, aminoacids are grouped into blocks of this '
                           'size '
                           'that are moved translationally and rotationally together.')

        form.addSection(label='Animation')
        form.addParam('amplitude', FloatParam, default=50,
                      label='Amplitude',
                      help='Used only for animations of computed normal modes. '
                           'This is the amplitude with which atoms or pseudoatoms are moved '
                           'along normal modes in the animations. \n'
                           'Normal-mode amplitudes corresponding to given images are computed by image analysis.')
        form.addParam('nframes', IntParam, default=10,
                      expertLevel=LEVEL_ADVANCED,
                      label='Number of frames',
                      help='Number of frames used in animations.')
        form.addParam('downsample', FloatParam, default=1,
                      expertLevel=LEVEL_ADVANCED,
                      # condition=isEm
                      label='Downsample pseudoatoms (for visualization)',
                      help='Used only with pseudoatoms and only for visualization purposes. \n'
                           'A downsample factor of 2 means removing one half of the pseudoatoms.')
        form.addParam('pseudoAtomThreshold', FloatParam, default=0,
                      expertLevel=LEVEL_ADVANCED,
                      # condition=isEm
                      label='Pseudoatom mass threshold (for visualization)',
                      help='Used only with pseudoatoms and only for visualization purposes. \n '
                           'Pseudoatoms whose mass is below this threshold are removed. \n'
                           'The threshold value should be between 0 and 1. '
                           'A threshold of 0 implies no pseudoatom removal.')

    def _insertAllSteps(self):
        # Some steps will differ if the input is a volume or a pdb file
        self.structureEM = self.inputStructure.get().getPseudoAtoms()
        n = self.numberOfModes.get()
        # Link the input
        inputFn = self.inputStructure.get().getFileName()
        localFn = self._getPath(replaceBaseExt(basename(inputFn), 'pdb'))
        self._insertFunctionStep('copyPdbStep', inputFn, localFn,
                                 self.structureEM)

        # Construct string for relative-absolute cutoff
        # This is used to detect when to reexecute a step or not
        cutoffStr = ''
        if self.cutoffMode == NMA_CUTOFF_REL:
            cutoffStr = 'Relative %f' % self.rcPercentage.get()
        else:
            cutoffStr = 'Absolute %f' % self.rc.get()

        # Compute modes
        self.pseudoAtomRadius = 1
        if self.structureEM:
            with open(inputFn, 'r') as fh:
                first_line = fh.readline()
                second_line = fh.readline()
                self.pseudoAtomRadius = float(second_line.split()[2])
            if self.cutoffMode == NMA_CUTOFF_REL:
                params = '-i %s --operation distance_histogram %s' \
                         % (localFn, self._getExtraPath('pseudoatoms_distance.hist'))
                self._insertFunctionStep('analyzePdbStep', params)

            self._insertFunctionStep('computeModesStep', localFn, n, cutoffStr)
            self._insertFunctionStep('reformatOutputStep', "pseudoatoms.pdb")
        else:
            if self.cutoffMode == NMA_CUTOFF_REL:
                params = '-i %s --operation distance_histogram %s' % (
                localFn, self._getExtraPath('atoms_distance.hist'))
                self._insertFunctionStep('analyzePdbStep', params)

            self._insertFunctionStep('computePdbModesStep', n,
                                     self.rtbBlockSize.get(),
                                     cutoffStr)
            self._insertFunctionStep('reformatPdbOutputStep', n)

        self._insertFunctionStep('qualifyModesStep', n,
                                 self.collectivityThreshold.get(),
                                 self.structureEM)
        self._insertFunctionStep('animateModesStep', n,
                                 self.amplitude.get(), self.nframes.get(),
                                 self.downsample.get(),
                                 self.pseudoAtomThreshold.get(),
                                 self.pseudoAtomRadius)
        self._insertFunctionStep('computeAtomShiftsStep', n)
        self._insertFunctionStep('createOutputStep')

    def copyPdbStep(self, inputFn, localFn, isEM):
        """ Copy the input pdb file and also create a link 'atoms.pdb'
        """
        if inputFn.endswith(".cif") or inputFn.endswith(".mmcif"):
            cifToPdb(inputFn, localFn)
        else:
            copyFile(inputFn, localFn)

        if isEM:
            fnOut = self._getPath('pseudoatoms.pdb')
        else:
            fnOut = self._getPath('atoms.pdb')

        if not os.path.exists(fnOut):
            createLink(localFn, fnOut)

        # Keeping only the lines that start with ATOM
        newlines = []
        with open(localFn) as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith("ATOM ") or line.startswith("TER ") or line.startswith("END "):
                newlines.append(line)
        with open(localFn, mode='w') as f:
            f.writelines(newlines)

        # Shifting the atom numbers after line 100000 one step to the left:
        newlines = []
        with open(localFn) as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith("ATOM ") or line.startswith("TER "):
                # print(int(line.split()[1]))
                if int(line.split()[1]) > 99999:
                    if line.startswith("ATOM "):
                        newline = line.replace("ATOM  1", "ATOM 1")
                    else:
                        newline = line.replace("TER   1", "TER  1")
                    newlines.append(newline)
                else:
                    newlines.append(line)
        with open(localFn, mode='w') as f:
            f.writelines(newlines)

    def analyzePdbStep(self, params):
        self.runJob("xmipp_pdb_analysis", params, env=Domain.importFromPlugin('xmipp3').Plugin.getEnviron())


    def computePdbModesStep(self, numberOfModes, RTBblockSize, cutoffStr):
        rc = self._getRc(self._getExtraPath('atoms_distance.hist'))

        self._enterWorkingDir()
        # For atoms, the interaction force constant was set to 10 as ElNemo RTB code may ask for its value \
        # (the RTBForceConstant entry was removed from gui as the value does not change the ENM computed normal modes).
        self.runJob('nma_record_info_PDB.py', "%d %d atoms.pdb %f %f"
                    % (numberOfModes, RTBblockSize, rc, 10.0),
                    env=getNMAEnviron())
        self.runJob("nma_elnemo_pdbmat", "", env=getNMAEnviron())
        self.runJob("nma_diagrtb", "", env=getNMAEnviron())

        if not exists("diagrtb.eigenfacs"):
            msg = "Modes cannot be computed. Check the number of modes you " \
                  "asked to compute and/or consider "
            msg += "increasing cut-off distance. The maximum number of " \
                   "modes allowed by the method for atomic "
            msg += "normal mode analysis is 6 times the number of RTB blocks " \
                   "but the protocol allows only up "
            msg += "to 200 modes as 20-100 modes are usually enough. If the " \
                   "number of modes is below the minimum "
            msg += "between 200 and 6 times the number of RTB blocks, consider " \
                   "increasing cut-off distance."
            self._printWarnings(redStr(msg) + '\n')
        self.runJob("rm", "-f *.dat_run diagrtb.dat pdbmat.xyzm pdbmat.sdijf "
                          "pdbmat.dat")

        self._leaveWorkingDir()

    def reformatPdbOutputStep(self, numberOfModes):
        self._enterWorkingDir()

        makePath('modes')
        Natoms = self._countAtoms("atoms.pdb")
        fhIn = open('diagrtb.eigenfacs')
        fhAni = open('vec_ani.txt', 'w')

        for n in range(numberOfModes):
            # Skip two lines
            fhIn.readline()
            fhIn.readline()
            fhOut = open('modes/vec.%d' % (n + 1), 'w')
            for i in range(Natoms):
                line = fhIn.readline()
                fhOut.write(line)
                fhAni.write(line.rstrip().lstrip() + " ")
            fhOut.close()
            if n != (numberOfModes - 1):
                fhAni.write("\n")
        fhIn.close()
        fhAni.close()
        self.runJob("nma_prepare_for_animate.py", "", env=getNMAEnviron())
        cleanPath("vec_ani.txt")
        moveFile('vec_ani.pkl', 'extra/vec_ani.pkl')

        self._leaveWorkingDir()

    def animateModesStep(self, numberOfModes, amplitude, nFrames, downsample,
                         pseudoAtomThreshold, pseudoAtomRadius):
        makePath(self._getExtraPath('animations'))
        self._enterWorkingDir()

        if self.structureEM:
            fn = "pseudoatoms.pdb"
            self.runJob("nma_animate_pseudoatoms.py", "%s extra/vec_ani.pkl 7 %d "
                                                      "%f extra/animations/"
                                                      "animated_mode %d %d %f" % \
                        (fn, numberOfModes, amplitude, nFrames, downsample,
                         pseudoAtomThreshold), env=getNMAEnviron())
        else:
            fn = "atoms.pdb"
            self.runJob("nma_animate_atoms.py", "%s extra/vec_ani.pkl 7 %d %f "
                                                "extra/animations/animated_mode "
                                                "%d" % \
                        (fn, numberOfModes, amplitude, nFrames), env=getNMAEnviron())

        for mode in range(7, numberOfModes + 1):
            fnAnimation = join("extra", "animations", "animated_mode_%03d"
                               % mode)
            fhCmd = open(fnAnimation + ".vmd", 'w')
            fhCmd.write("mol new %s.pdb\n" % self._getPath(fnAnimation))
            fhCmd.write("animate style Loop\n")
            fhCmd.write("display projection Orthographic\n")
            if self.structureEM:
                fhCmd.write("mol modcolor 0 0 Beta\n")
                fhCmd.write("mol modstyle 0 0 Beads %f 8.000000\n"
                            % (pseudoAtomRadius))
            else:
                fhCmd.write("mol modcolor 0 0 Index\n")
                if self._checkPDB_CA(fn):
                    fhCmd.write("mol modstyle 0 0 Beads 1.000000 8.000000\n")
                    # fhCmd.write("mol modstyle 0 0 Beads 1.800000 6.000000 "
                    #         "2.600000 0\n")
                else:
                    fhCmd.write("mol modstyle 0 0 NewRibbons 1.800000 6.000000 "
                                "2.600000 0\n")
            fhCmd.write("animate speed 0.5\n")
            fhCmd.write("animate forward\n")
            fhCmd.close();

        self._leaveWorkingDir()

    def computeAtomShiftsStep(self, numberOfModes):
        fnOutDir = self._getExtraPath("distanceProfiles")
        makePath(fnOutDir)
        maxShift = []
        maxShiftMode = []

        for n in range(7, numberOfModes + 1):
            fnVec = self._getPath("modes", "vec.%d" % n)
            if exists(fnVec):
                fhIn = open(fnVec)
                md = MetaData()
                atomCounter = 0
                for line in fhIn:
                    x, y, z = map(float, line.split())
                    d = math.sqrt(x * x + y * y + z * z)
                    if n == 7:
                        maxShift.append(d)
                        maxShiftMode.append(7)
                    else:
                        if d > maxShift[atomCounter]:
                            maxShift[atomCounter] = d
                            maxShiftMode[atomCounter] = n
                    atomCounter += 1
                    md.setValue(MDL_NMA_ATOMSHIFT, d, md.addObject())
                md.write(join(fnOutDir, "vec%d.xmd" % n))
                fhIn.close()
        md = MetaData()
        for i, _ in enumerate(maxShift):
            fnVec = self._getPath("modes", "vec.%d" % (maxShiftMode[i] + 1))
            if exists(fnVec):
                objId = md.addObject()
                md.setValue(MDL_NMA_ATOMSHIFT, maxShift[i], objId)
                md.setValue(MDL_NMA_MODEFILE, fnVec, objId)
        md.write(self._getExtraPath('maxAtomShifts.xmd'))

    def createOutputStep(self):
        fnSqlite = self._getPath('modes.sqlite')
        nmSet = SetOfNormalModes(filename=fnSqlite)

        md = MetaData(self._getPath('modes.xmd'))
        row = XmippMdRow()

        for objId in md:
            row.readFromMd(md, objId)
            nmSet.append(rowToMode(row))
        inputPdb = self.inputStructure.get()
        nmSet.setPdb(inputPdb)
        self._defineOutputs(outputModes=nmSet)
        self._defineSourceRelation(self.inputStructure, nmSet)

    def _checkPDB_CA(self, fnPDB):
        # This function returns true if all the atoms are CA and P, otherwise false
        from continuousflex.protocols.utilities.pdb_parser import m_inout_read_pdb
        pdb_read = m_inout_read_pdb(fnPDB)
        for atom in pdb_read:
            if atom.type != " C" or atom.loc != "A ":
                if atom.type != " P":
                    return False
        return True

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass


import numpy as np
from sklearn.preprocessing import StandardScaler

def read_pdb(path, ca = False):
    with open(path, 'r') as f:
        lines = f.readlines()
    pdb_list = []
    for line in lines:
        line = line[:38] + " " + line[38:]
        line = line[:47] + " " + line[47:]
        line = line[:62] + " " + line[62:]
        if ca:
            if line.startswith("ATOM") and " CA " in line:
                line = line.split()
                pdb_list.append(line[6:9])
            else:
                pass
        else:
            if line.startswith("ATOM"):
                line = line.split()
                pdb_list.append(line[6:9])
            else:
                pass

    # pythonic way
    #lines = [line.split for line in lines]
    #lines = [line for line in lines if "ATOM" in line]
    #if ca:
    #    lines = [line for line in lines if " CA " in line]
    #lines
    pdb_array = np.array(pdb_list, dtype='float32')
    coords = pdb_array
    return coords
    
def parse_pdb(path, atom_p, i):
    """
    ----------
    This function generates the predicted pdb.

    Parameters
    ----------
    path : string path to the reference structure.
    atom_p: numpy array of the predicted atom positions.

    Returns
    -------
    """
    with open(path, 'r') as f:
            lines = f.readlines()
    lines = [line for line in lines if 'ATOM' in line]
    atom = [line[0:4] for line in lines]
    serial = ['{:>7}'.format(line[4:12].strip()) for line in lines]
    atom_name = ['{:>4}'.format(line[12:17].strip()) for line in lines]
    loc_ind = ['{:>5}'.format(line[17:20].strip()) for line in lines]
    residue_name = ['{:>2}'.format(line[21:22].strip()) for line in lines]
    chain_id = ['{:>4}'.format(line[23:28].strip()) for line in lines]
    occupancy = ['{:>6}'.format(line[56:60].strip()) for line in lines]
    temperature = ['{:>6}'.format(line[60:67].strip()) for line in lines]
    elem_symb = ['{:>12}'.format(line[77:80].strip()) for line in lines]
    with open('synth'+str(i)+'.pdb','w') as f:
        for i in range(len(lines)):
            pos = '{:>12.6}{:>8.6}{:>8.5}'.format(atom_p[i,0],atom_p[i,1],atom_p[i,2])
            line = atom[i]+serial[i]+atom_name[i]+loc_ind[i]+residue_name[i]+chain_id[i]+\
                   pos+occupancy[i]+temperature[i]+elem_symb[i]+'\n'
            f.write(line)
            
def standard_pdb(coords):
    return (coords-np.mean(coords,axis=(0)))/np.std(coords,axis=(0))



# **************************************************************************
# *
# * Authors:
# * Ilyes Hamitouche
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import xmipp3.convert
from pyworkflow.protocol.params import PointerParam
import pyworkflow.protocol.params as params
from pwem.protocols import ProtAnalysis3D
from subprocess import check_call
import sys
from os.path import isfile
import continuousflex
from pyworkflow.utils.path import copyFile
import pwem as em
import pwem.emlib.metadata as md
from xmipp3.convert import (createItemMatrix, setXmippAttributes)
from pyworkflow import BETA
from continuousflex import Plugin

OPTION_NMA = 0
OPTION_ANGLES = 1
OPTION_SHFITS = 2
OPTION_ALL = 3

DEVICE_CUDA = 0
DEVICE_CPU = 1

class FlexProtDeepHEMNMAInfer(ProtAnalysis3D):
    """ This protocol is DeepHEMNMA
    """
    _label = 'deep hemnma infer'
    _devStatus = BETA

    
    #--------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('analyze_option', params.EnumParam, label='set the parameter to predict',
                      display=params.EnumParam.DISPLAY_COMBO,
                      choices=['Predict Normal Mode Amplitudes',
                               'Predict Angles',
                               'Predict Shifts',
                               'predict All parameters',
                               ], default=OPTION_ALL,
                      help='select a set of parameter to predict')
        form.addParam('device_option', params.EnumParam, label='set the device for training',
                      display=params.EnumParam.DISPLAY_COMBO,
                      choices=['train on GPUs',
                               'tain on CPUs'], default=DEVICE_CUDA,
                      help='set a device to run the training on')
        form.addParam('trained_model', params.PointerParam, pointerClass='FlexProtDeepHEMNMATrain',
                      label = 'Trained model', help='import the training weights')
        form.addParam('inputParticles', PointerParam, pointerClass='SetOfParticles',
                       label="Previous run of rigid-body alignment",
                       help='Select a previous run of rigid-body alignment.', allowsNull=True)
        form.addParam('num_modes', params.IntParam, label='Number of modes',default=3)
        form.addParam('batch_size', params.IntParam, expertLevel=params.LEVEL_ADVANCED, label='Batch size', default=2)
        form.addParallelSection(threads=0, mpi=0)    
    
    
    #--------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        self._insertFunctionStep('convertInputStep')
        self._insertFunctionStep('performDeepHEMNMAStep')
        self._insertFunctionStep('createOutputStep')
        
    #--------------------------- STEPS functions --------------------------------------------   
    
    def convertInputStep(self):
        xmipp3.convert.writeSetOfParticles(self.inputParticles.get(), self._getExtraPath('particles.xmd'))
        # copy atoms or pseudoatoms file from HEMNMA
        file = self.trained_model.get().inputNMA.get()._getExtraPath('atoms.pdb')
        if isfile(file):
            copyFile(file, self._getExtraPath('atoms.pdb'))
        else:
            copyFile(self.trained_model.get().inputNMA.get()._getExtraPath('pseudoatoms.pdb'), self._getExtraPath('pseudoatoms.pdb'))


    
    def performDeepHEMNMAStep(self):
        weights = self.trained_model.get()._getExtraPath('weights.pth')
        #copyFile(self.inputParticles.get('atoms.pdb'), self._getExtraPath('atoms.pdb'))
        #copyFile(self.inputParticles.get('modes.pdb'), self._getExtraPath('modes.pdb'))
        batch_size = self.batch_size.get()
        mode = self.analyze_option.get()
        device = self.device_option.get()
        num_modes = self.num_modes.get()
        self.imgsFn = self._getExtraPath('particles.xmd')
        params = " %s %s %s %d %d %d %d" % (self.imgsFn, weights, self._getExtraPath(), num_modes, batch_size, mode, device)
        script_path = continuousflex.__path__[0]+'/protocols/utilities/deep_hemnma_infer.py'
        command = "python " + script_path + params
        command = Plugin.getContinuousFlexCmd(command)
        check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr, env=None, cwd=None)
        pass

        
    def createOutputStep(self):
        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticles()
        partSet.copyInfo(inputSet)
        self.imgsFn = self._getExtraPath('images.xmd')
        copyFile(self.imgsFn, self._getExtraPath('infer.xmd'))
        partSet.copyItems(inputSet,
                          updateItemCallback=self._updateParticle,
                          itemDataIterator=md.iterRows(self.imgsFn, sortByLabel=md.MDL_ITEM_ID))
        partSet.copyItems(self.trained_model.get().inputNMA.get().outputParticles)
        self._defineOutputs(outputParticles=partSet)
        # Lets write a metadata that combines both of these training and inference sets:
        fn_train = self.trained_model.get().inputNMA.get()._getExtraPath('images.xmd')
        fn_infer = self._getExtraPath('infer.xmd')
        fn_combined = self._getExtraPath('images.xmd')
        args = '-i %(fn_train)s -o %(fn_combined)s --set union %(fn_infer)s' % locals()
        self.runJob('xmipp_metadata_utilities', args)


    #--------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary
    
    def _validate(self):
        errors = []
        return errors
    
    def _citations(self):
        return ['harastani2022continuousflex','hamitouche2022deephemnma']
    
    def _methods(self):
        return []
    
    #--------------------------- UTILS functions --------------------------------------------

    def getInputParticles(self):
        """ Get the output particles of the input NMA protocol. """
        return self.inputNMA.get().outputParticles

    def getParticlesMD(self):
        "Get the metadata files that contain the NMA displacement"
        return self.inputNMA.get()._getExtraPath('images.xmd')

    def getInputPdb(self):
        return self.inputNMA.get().getInputPdb()
    
    def getOutputMatrixFile(self):
        return self._getExtraPath('output_matrix.txt')
    
    def getDeformationFile(self):
        return self._getExtraPath('deformations.txt')
    
    def getProjectorFile(self):
        return self.mappingFile.get()


    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT, md.MDL_ANGLE_PSI, md.MDL_SHIFT_X,
                           md.MDL_SHIFT_Y, md.MDL_FLIP, md.MDL_NMA, md.MDL_COST)
        createItemMatrix(item, row, align=em.ALIGN_PROJ)


# By Mohamad Harastani

from spider_files3 import save_volume
import sys
import farneback3d
import joblib

def opflow_warp(ref_dump, flow_dump, warped_path_i):
    reference = joblib.load(ref_dump)
    flow_i = joblib.load(flow_dump)
    warped_i = farneback3d.warp_by_flow(reference, flow_i)
    save_volume(warped_i, warped_path_i)


if __name__ == '__main__':
    opflow_warp(sys.argv[1],
                sys.argv[2],
                sys.argv[3])
    sys.exit()


import torch.nn as nn

class mlp(nn.Module):
    def __init__(self, output):
        super(mlp, self).__init__()

        hidden_dims = [512,1000, 512, 128]
        modules = []
        for i in range(len(hidden_dims)-1):
            modules.append(nn.Sequential(nn.Linear(hidden_dims[i], hidden_dims[i+1]),
                                         nn.LayerNorm(hidden_dims[i+1]),
                                         nn.GELU(),
                                         
                                         ))
        self.mlp_ = nn.Sequential(*modules)

        self.output_layer = nn.Linear(hidden_dims[-1], output)

    def forward(self, x):
        mlp_ = self.mlp_(x)
        elu = nn.ELU()
        output_layer = elu(self.output_layer(mlp_))
        return output_layer



# **************************************************************************
# * Authors:
# * Mohamad Harastani (mohamad.harastani@igbmc.fr)
# * Slavica Jonic (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from setuptools import setup, find_packages
from codecs import open
from os import path
from continuousflex import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='scipion-em-continuousflex',
    version=__version__,
    description='Plugin to use continuousflex protocols within the Scipion framework',
    long_description=long_description,
    url='https://github.com/scipion-em/scipion-em-continuousflex',
    author='Mohamad Harastani, Remi Vuillemot, Ilyes Hamitouche and Slavica Jonic',
    author_email='slavica.jonic@upmc.fr',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3'
    ],

    keywords='cryo-em cryo-et image-processing continuous-conformational-variability',  # Optional
    packages=find_packages(),
    install_requires=[requirements],
    include_package_data=True,
    package_data={
       'continuousflex': ['protocols.conf'],
    },

    entry_points={
        'pyworkflow.plugin': 'continuousflex = continuousflex'
    },

    project_urls={
        'Bug Reports': 'https://github.com/scipion-em/scipion-em-continuousflex/issues',
        'Source': 'https://github.com/scipion-em/scipion-em-continuousflex/',
    },
)



from .metadata import read_file
from .metadata import min_max, standardization, reverse_min_max, reverse_standardization
from .spi_reader import spi2array, normalize, torch_normalize
from .spi_reader import read_from_list, read_from_directory
from .pdb_reader import read_pdb, parse_pdb
from .euler2quaternion import eul2quat, quater2euler
from .projection import projectPDB2Image
from .projection import projectPDB_NP



#!/usr/bin/env python3
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
# m_inout.py  - Library for coordinate input and output
# Part of the ModeHunter package, http://modehunter.biomachina.org
#
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
# Request to cite published literature:
#
# When using this software in scholarly work please cite the publications(s)
# posted at http://modehunter.biomachina.org/fref.html
#
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
# Legal notice:
#
# This software is copyrighted, (c) 2009, by Joseph N. Stember and Willy Wriggers
# under the following terms:
#
# The authors hereby grant permission to use, copy, modify, and re-distribute this
# software and its documentation for any purpose, provided that existing copyright
# notices are retained in all copies and that this notice is included verbatim in
# any distributions. No written agreement, license, or royalty fee is required for
# any of the authorized uses.
#
# IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR DIRECT,
# INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE
# OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY DERIVATIVES THEREOF, EVEN IF THE
# AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON AN "AS
# IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE
# MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy
import sys
import copy

# global list with chemical element names
element_name = ["H", "HE", "LI", "BE", "B", "C", "N", "O", "F", "NE",
                "NA", "MG", "AL", "SI", "P", "S", "CL", "AR", "K", "CA",
                "SC", "TI", "V", "CR", "MN", "FE", "CO", "NI", "CU", "ZN"]

# global list with corresponding atom masses
element_mass = [1.008, 4.003, 6.940, 9.012, 10.810, 12.011, 14.007, 16.000, 18.998, 20.179,
                22.990, 24.305, 26.982, 28.086, 30.974, 32.060, 35.453, 39.948, 39.102, 40.080,
                44.956, 47.880, 50.942, 51.996, 54.938, 55.847, 58.933, 58.700, 63.546, 65.380]


# global PDB class with data required structure
class PDB(object):
    def __init__(self, recd="ATOM  ", serial=1, type="  ", loc="  ", alt=" ", res="    ", chain=" ", seq=1, icode=" ",
                 x=0.0, y=0.0, z=0.0, occupancy=0.0, beta=0.0, footnote=0, segid="    ", element="  ", charge="  ",
                 weight=0.0):
        self.recd = recd  # 1- 6
        self.serial = serial  # 7-11
        self.type = type  # 13-14
        self.loc = loc  # 15-16
        self.alt = alt  # 17
        self.res = res  # 18-21
        self.chain = chain  # 22
        self.seq = seq  # 23-26
        self.icode = icode  # 27
        self.x = x  # 31-38
        self.y = y  # 39-46
        self.z = z  # 47-54
        self.occupancy = occupancy  # 55-60
        self.beta = beta  # 61-66
        self.footnote = footnote  # 68-70
        self.segid = segid  # 73-76
        self.element = element  # 77-78
        self.charge = charge  # 79-80
        self.weight = weight  # mass of atom assigned by read_pdb


# functions

def m_inout_read_pdb(filename):
    "Situs 2.x style pdb parser"

    pdb_list = []
    f = open(filename, 'r')
    for line in f:
        test_recd = line[0:5]

        # if test_recd.upper().strip() == "ATOM" or test_recd.upper() == "HETATM":
        if ((test_recd.upper().strip() == "ATOM") or (test_recd.upper().strip() == "HETATM")):
            # create new atom and assign standard fields
            new_atom = PDB()
            new_atom.recd = test_recd
            try:
                new_atom.serial = int(line[4:11])
            except:
                pass
            try:
                new_atom.type = line[12:14]
            except:
                pass
            try:
                new_atom.loc = line[14:16]
            except:
                pass
            try:
                new_atom.alt = line[16]
            except:
                pass
            try:
                new_atom.res = line[17:21]
            except:
                pass
            try:
                new_atom.chain = line[21]
            except:
                pass
            try:
                new_atom.seq = int(line[22:26])
            except:
                pass
            try:
                new_atom.icode = line[26]
            except:
                pass
            try:
                new_atom.x = float(line[30:38])
            except:
                pass
            try:
                new_atom.y = float(line[38:46])
            except:
                pass
            try:
                new_atom.z = float(line[46:54])
            except:
                pass
            try:
                new_atom.occupancy = float(line[54:60])
            except:
                pass
            try:
                new_atom.beta = float(line[60:66])
            except:
                pass
            try:
                new_atom.footnote = int(line[67:70])
            except:
                pass
            try:
                new_atom.segid = line[72:76]
            except:
                pass
            try:
                new_atom.element = line[76:78]
            except:
                pass
            try:
                new_atom.charge = line[78:80]
            except:
                pass
            # compute robust Situs 2.x style mass weight and append atom
            if (new_atom.type == "QV" and new_atom.loc == "OL") or (new_atom.type == "QP" and new_atom.loc == "DB"):
                new_atom.weight = 0.0  # codebook vectors have zero mass
            elif (new_atom.type == "DE" and new_atom.loc == "NS"):
                new_atom.weight = new_atom.occupancy  # volumetric map density mass assigned from occupancy field
            elif (new_atom.type.lstrip()[0] == "H"):
                new_atom.weight = element_mass[0]  # quick hack for hydrogens, may return false positives for Hg, Hf, Ho
            elif (new_atom.type.strip in element_name):
                new_atom.weight = element_mass[element_name.index(new_atom.type.strip)]  # standard element
            elif (new_atom.type.strip()[0] in element_name):
                new_atom.weight = element_mass[
                    element_name.index(new_atom.type.strip()[0])]  # catches frequent elements, C N O P S
            else:
                # Slavica: Mass of other atoms (eg, 1H,2H,3H, 1HH1, 2HH1, 1HD2, 2HD2 etc) to be incorporated later
                # print "Warning: atom %i: unable to identify atom type %s, assigning carbon mass" % (len(pdb_list)+1,new_atom.type)
                new_atom.weight = element_mass[5]
            pdb_list.append(new_atom)

    f.close()
    return pdb_list


def m_inout_write_pdb(pdb_list, filename, remarks):
    "pdb writer, Situs 2.x convention, renumbering atoms"

    f = open(filename, 'w')
    f.write("REMARKS " + remarks + "\n")

    for i in range(len(pdb_list)):
        f.write(pdb_list[i].recd.strip()[:6].ljust(6))
        # renumber atoms, ignoring .serial
        if ((i + 1) // 100000 == 0):
            f.write("%5d" % (i + 1))
        else:
            f.write("%05d" % (i + 1 - ((i + 1) // 100000) * 100000))
        f.write(" ")
        f.write(pdb_list[i].type.strip()[:2].rjust(2))
        f.write(pdb_list[i].loc.strip()[:2].ljust(2))
        f.write(pdb_list[i].alt.strip()[:1].ljust(1))
        if (len(pdb_list[i].res) < 4):
            f.write(pdb_list[i].res.strip().rjust(3))
            f.write(" ")
        else:
            f.write(pdb_list[i].res[:4])
        f.write(pdb_list[i].chain.strip()[:1].ljust(1))
        f.write("%4d" % pdb_list[i].seq)
        f.write(pdb_list[i].icode.strip()[:1].ljust(1))
        f.write("   ")
        f.write("%8.3f" % pdb_list[i].x)
        f.write("%8.3f" % pdb_list[i].y)
        f.write("%8.3f" % pdb_list[i].z)
        f.write("%6.2f" % pdb_list[i].occupancy)
        f.write("%6.2f" % pdb_list[i].beta)
        f.write(" ")
        f.write("%3d" % pdb_list[i].footnote)
        f.write("  ")
        f.write(pdb_list[i].segid.strip()[:4].ljust(4))
        f.write(pdb_list[i].element.strip()[:2].rjust(2))
        f.write(pdb_list[i].charge.strip()[:2].rjust(2))
        f.write("\n")
    f.close()


def m_inout_write_pdb_sampled(pdb_name, filename, step, thr_mass):
    "pdb writer, Situs 2.x convention, renumbering atoms"

    f = open(filename, 'w')

    pdb_list = m_inout_read_pdb(pdb_name)

    for i in range(0, len(pdb_list), step):
        if pdb_list[i].beta > thr_mass:
            f.write(pdb_list[i].recd.strip()[:6].ljust(6))
            # renumber atoms, ignoring .serial
            if ((i + 1) // 100000 == 0):
                f.write("%5d" % (i + 1))
            else:
                f.write("%05d" % (i + 1 - ((i + 1) // 100000) * 100000))
            f.write(" ")
            f.write(pdb_list[i].type.strip()[:2].rjust(2))
            f.write(pdb_list[i].loc.strip()[:2].ljust(2))
            f.write(pdb_list[i].alt.strip()[:1].ljust(1))
            if (len(pdb_list[i].res) < 4):
                f.write(pdb_list[i].res.strip().rjust(3))
                f.write(" ")
            else:
                f.write(pdb_list[i].res[:4])
            f.write(pdb_list[i].chain.strip()[:1].ljust(1))
            f.write("%4d" % pdb_list[i].seq)
            f.write(pdb_list[i].icode.strip()[:1].ljust(1))
            f.write("   ")
            f.write("%8.3f" % pdb_list[i].x)
            f.write("%8.3f" % pdb_list[i].y)
            f.write("%8.3f" % pdb_list[i].z)
            f.write("%6.2f" % pdb_list[i].occupancy)
            f.write("%6.2f" % pdb_list[i].beta)
            f.write(" ")
            f.write("%3d" % pdb_list[i].footnote)
            f.write("  ")
            f.write(pdb_list[i].segid.strip()[:4].ljust(4))
            f.write(pdb_list[i].element.strip()[:2].rjust(2))
            f.write(pdb_list[i].charge.strip()[:2].rjust(2))
            f.write("\n")
    f.close()


def m_inout_carbon_alphas(pdb_list):
    "return 0-based indices of carbon alphas in pdb_list"

    cas = []
    for i in range(len(pdb_list)):
        if (pdb_list[i].type == " C" and pdb_list[i].loc == "A "):
            cas.append(i)
    return cas


def m_inout_import_coords(filename):
    "get x y z coords from PDB file"
    coords = numpy.array([])
    pdb_list = m_inout_read_pdb(filename)
    for i in numpy.arange(len(pdb_list)):
        coords = numpy.append(coords, [pdb_list[i].x, pdb_list[i].y, pdb_list[i].z])
    return coords


def m_inout_import_bfact(filename):
    "get Bfactor from PDB file"
    bfact = numpy.array([])
    pdb_list = m_inout_read_pdb(filename)
    for i in numpy.arange(len(pdb_list)):
        bfact = numpy.append(bfact, [pdb_list[i].beta])
    return bfact